GDC.charting.data.MDataSet = function (dataSetInput) {
    this.dataSet = new GDC.grid.DataSet(dataSetInput);
    ds = this;
    this.dataSet.subscribe('onload', function() {
        this.attributes = {
            columns: undefined,
            rows: undefined,
            metrics: undefined
        };
        this.metricIds = [];
        
        this.convertAttributeLevelsToSeries(this.dataSet.getReportDefinition());
        GDC.util.log("Got attribute levels");
        this.attributesPerSection = this.bindAtributesToControlledSeries();
        GDC.util.log("Bind them to series");
        var metricsInRows = this.getReportDefinition().getMetricGroupPosition().section === "rows";
        this.attributeSeries = metricsInRows ? this.attributesPerSection.rows.concat(this.attributesPerSection.columns) : this.attributesPerSection.columns.concat(this.attributesPerSection.rows);
        GDC.util.log("Building the tree");
        this.buildDataIndexTree();
        
        return;
    }, this, true);
    
    
    this.subscribe("dataDownloadingFinished", function(data){
        this.metricData = data;
        this.dataSeries = this.bindDataTreeToControlledSeries(this.metricData);
        db = new GDC.charting.data.DataBag(this.attributesPerSection, this.dataSeries);
        this.fireEvent("ready", db);
    }, this, true);
    this.createEvent("ready");
    this.createEvent("dataDownloadingFinished");
};

YAHOO.lang.augmentObject(GDC.charting.data.MDataSet.prototype, {
    getReportDefinition: function() {
        return this.dataSet.getReportDefinition();
    },
    
    convertAttributeLevelsToSeries: function(reportDefinition) {
        var interestingLevels;
        var numberOfColumnAttributes;
        var numberOfRowAttributes;
        
        numberOfColumnAttributes = reportDefinition.getColAttributes().length-1;
        interestingColumnLevels = GDC.util.interval(0, numberOfColumnAttributes);
        this.attributes.columns = (this.getAttributesOnLevel(this.dataSet.getColumnHeaders(), interestingColumnLevels));
        
        if (reportDefinition.getMetricGroupPosition().section === "rows") {
            this.attributes.metrics = (this.getAttributesOnLevel(this.dataSet.getRowHeaders(), [reportDefinition.getMetricGroupPosition().index]));
        } else {
            this.attributes.metrics = (this.getAttributesOnLevel(this.dataSet.getColumnHeaders(), [reportDefinition.getMetricGroupPosition().index]));
        }
        
        numberOfRowAttributes = reportDefinition.getRowAttributes().length-1;
        interestingRowLevels = GDC.util.interval(0, numberOfRowAttributes);
        this.attributes.rows = (this.getAttributesOnLevel(this.dataSet.getRowHeaders(), interestingRowLevels));
        
        this.metricIds = GDC.util.array.unique(this.metricIds);
    },
    
    getAttributesOnLevel: function(treeContainer, levels) {
        var lookups = treeContainer.lookups;
        var attributes = [];
        levels.forEach(function(level) {
            attributes.push(this.traverseTreeDF(treeContainer.tree, -1, function(node, level) {
                var treeNode = new GDC.charting.data.TreeNode({
                    id: node.id,
                    type: node.type,
                    label: treeContainer.lookups[level][node.id],
                    distance: node.first
                });
                if (treeNode.getType() === "metric") this.metricIds.push(treeNode.getId());
                return treeNode;
            }, this, level));
        }, this);
        return new GDC.charting.data.Series(attributes);
    },
    
    
    
    /**
     * @method getAttribute
     * Binding atttributes together
     */
    bindAtributesToControlledSeries: function() {
        var reportDefinition = this.dataSet.getReportDefinition();
        var columnLevels = [];
        this.attributes.columns.map(function(attributeLevel) { return attributeLevel.getLabel();}).forEach(function(attributeLevel, index) {
            var data;
            // GDC.util.log("Level", attributeLevel);
            if (index === 0) {
                // var series = new GDC.charting.data.Series(attributeLevel);
                data = new GDC.charting.data.IteratingSeries(attributeLevel);
            } else {
                data = new GDC.charting.data.IteratingSeries(new GDC.charting.data.ControlledSeries(attributeLevel, columnLevels[index-1].data));
            }
            columnLevels.push(this.createStructure({
                type:       "attribute",
                name:       reportDefinition.getColAttributeByIndex(index).getTitle(),
                data:       data,
                section:    "columns"
            }));
        }, this);
        
        // var lastColumnsIndex = levels.length - 1;
        rowLevels = [];
        this.attributes.rows.map(function(attributeLevel) { return attributeLevel.getLabel();}).forEach(function(attributeLevel, index) {
            var data;
            if (index === 0) {
                // var series = new GDC.charting.data.Series(attributeLevel);
                data = new GDC.charting.data.IteratingSeries(attributeLevel);
            } else {
                data = new GDC.charting.data.IteratingSeries(new GDC.charting.data.ControlledSeries(attributeLevel, rowLevels[index - 1].data));
            }
            rowLevels.push(this.createStructure({
                type:       "attribute",
                name:       reportDefinition.getRowAttributeByIndex(index).getTitle(),
                data:       data,
                section:    "rows"
            }));
            
        }, this);
        
        // var lastRowsIndex = rowLevels.length - 1;
        // if (reportDefinition.getMetricGroupPosition().section === "rows") {
        //             this.attributes.metrics.map(function(attributeLevel) { return attributeLevel.getLabel();}).forEach(function(attributeLevel, index) {
        //                 var data = lastRowsIndex === -1 ? new GDC.charting.data.IteratingSeries(attributeLevel) : new GDC.charting.data.IteratingSeries(new GDC.charting.data.ControlledSeries(attributeLevel, rowLevels[lastRowsIndex + index].data));
        //                 rowLevels.push(this.createStructure({
        //                     type: "metric",
        //                     data: data
        //                 }));
        //             }, this);
        //         }
        return {
            rows: rowLevels,
            columns: columnLevels
        };
    },
    
    createStructure: function(data) {
        return {
            type:       data.type,
            name:       data.name,
            data:       data.data,
            section:    data.section
        };
    },
    
    getFirstStageData: function() {
        var reportDefinition = this.dataSet.getReportDefinition();
        var metricInRows = reportDefinition.getMetricGroupPosition().section === "rows";
        return metricInRows ? this.dataSet.getRowHeaders() : this.dataSet.getColumnHeaders();
    },
    
    getSecondStageData: function() {
        var reportDefinition = this.dataSet.getReportDefinition();
        var metricInRows = reportDefinition.getMetricGroupPosition().section === "rows";
        return metricInRows ? this.dataSet.getColumnHeaders() : this.dataSet.getRowHeaders();
    },
    
    buildDataIndexTree: function() {
        // this.attributes.metrics.map;
        var dataTree = [];
        this.numbers = 0;
        dt = dataTree;
        path = [];
        this.metricIds.forEach(function(metricId, index) {
            path.push(index);
            var x = this.firstStage(this.getFirstStageData().tree, -1, path, {
                stageCallback: function(firstStageResult) {
                    
                    return this.secondStage(this.getSecondStageData().tree, -1, path, {
                        stageCallback: function(point, path) {
                            // GDC.util.log(point.column, point.row);
                            // GDC.util.log("path", path);
                            // var dataSet = dataTree;
                            var setData = function(dataSet, data, path) {
                                var d = dataSet;
                                var lastIndex = path.pop(); 
                                // GDC.util.log(path);
                                path.forEach(function(pathIndex, index) {
                                    // GDC.util.log("pathindex", pathIndex, "leads to", d[pathIndex]);
                                    if (YAHOO.lang.isArray(d) && YAHOO.lang.isObject(d[0]) && d[0].metric) {
                                        d = d[0].metric;
                                    } else if (YAHOO.lang.isArray(d)) {
                                        d = d[pathIndex];
                                    } else {
                                        GDC.util.log("something else, should not be :-)", d);
                                    }
                                });
                                // GDC.util.log("+++++++++++");
                                // GDC.util.log(d);
                                // GDC.util.log(d[lastIndex]);
                                if (YAHOO.lang.isArray(d)) {
                                    d = d[lastIndex] = data;
                                } else {
                                    d = d.metric[lastIndex] = data;
                                }
                                // d[lastIndex] = data;
                                // GDC.util.log(d[lastIndex]);
                                // GDC.util.log("+++++++++++");
                                
                            };
                            this.dataSet.getCell(point.column + this.dataSet.getFrozenCells().left, point.row + this.dataSet.getFrozenCells().top, {
                                success: function(obj) {
                                    // GDC.util.log(path);
                                    // GDC.util.log("[", point.column,",",point.row,"]");
                                    // GDC.util.log(dataTree);
                                    // GDC.util.log(path);
                                    // GDC.util.log("++++ DATA VALUE ++++ ",obj.value);
                                    // GDC.util.log(dataTree);
                                    setData(dataTree, obj.value, GDC.util.clone(path));
                                    // GDC.util.log(dataTree);
                                    this.lastStage(dataTree);
                                },
                                scope: this
                            });
                        },
                        scope: this,
                        firstStageResult: firstStageResult,
                        metricId: metricId
                    });
                },
                scope: this,
                metricId: metricId
            }, this);
            path.pop();
            dataTree.push(x);
        }, this);
        return dataTree;
    },
    
    firstStage: function(node, level, path, options) {
        // GDC.util.log("first stage");
        // GDC.util.log("level", level);
        var stageCallback = options.stageCallback;
        var scope = options.scope;
        var metricId = options.metricId;
        
        var nodes = [];
        if (node.children.length > 0) {
            
            node.children.forEach(function(child, index) {
                path.push(index);
                try {
                    var lowerLevel = this.firstStage(child, level+1, path, options);
                    nodes.push(lowerLevel);
                } catch(e) {
                    GDC.util.log("UNWANTED metric met, not pushing into the array");
                }
                path.pop();
            }, this);
            // if (YAHOO.lang.isArray(nodes[0])) {
                // GDC.util.log("is array");
                return nodes;
            // } else {
                // GDC.util.log("metric somewhere lower in", nodes);
                // return nodes[0].metric;
            // }
        } else {
            if (node.type === "metric" && node.id !== metricId) {
                throw "UnwantedMetric";
                // return;
            } else {
                var secondStageResult = stageCallback.call(scope, node.first);
                if (node.type === "metric") {
                    return {"metric": secondStageResult};
                } else {
                    return secondStageResult;
                }
            }
        }
    },
    
    secondStage: function(node, level, path, options) {
        // GDC.util.log("second stage");
        // GDC.util.log("2level", level);
        var stageCallback = options.stageCallback;
        var scope = options.scope;
        var firstStageResult = options.firstStageResult;
        var metricId = options.metricId;
        
        var nodes = [];
        if (node.children.length > 0) {
            node.children.forEach(function(child, index) {
                // GDC.util.log("pushing", index);
                path.push(index);
                // GDC.util.log("length", path.length);
                // GDC.util.log("path", path);
                nodes.push(this.secondStage(child, level+1, path, options));
                path.pop();
            }, this);
            return nodes;
        } else {
            this.numbers += 1;
            GDC.util.log(this.numbers);
            var point = this.createCellPointerStrucutre(firstStageResult, node.first);
            GDC.util.log(point);
            var thirdStageResult = stageCallback.call(scope, point, GDC.util.clone(path));
            var res = "[" + firstStageResult + "," + node.first + "]";
            // GDC.util.log(res);
            return res;
        }
    },
    
    createCellPointerStrucutre: function(firstStageResult, secondStageResult) {
        var reportDefinition = this.dataSet.getReportDefinition();
        var metricInRows = reportDefinition.getMetricGroupPosition().section === "rows";
        return metricInRows ? {column: secondStageResult, row: firstStageResult} : {column: firstStageResult, row: secondStageResult};
    },
    
    lastStage: function(dataTree) {
        // GDC.util.log("last stage");
        prunnedDataTree = [];
        this.numbers -= 1;
        GDC.util.log(this.numbers);
        if (this.numbers === 0) {
            // GDC.util.log(dataTree);
            prunnedDataTree = this.traverseTreeForPruningMetricLevels(dataTree, 0);
            // GDC.util.log("Prunned Data");
            // GDC.util.dir(prunnedDataTree);
            this.fireEvent("dataDownloadingFinished", prunnedDataTree);
        } else {
            // GDC.util.log("NOT FNITO");
        }
        
    },
    
    traverseTreeForPruningMetricLevels: function(node, level) {
        // GDC.util.log(node);
        var nodes = [];
        if (YAHOO.lang.isArray(node) || node.metric) {
            
            node = node[0].metric ? node[0].metric : node;
                
            node.forEach(function(child, index) {
                var lowerLevel = this.traverseTreeForPruningMetricLevels(child, level+1);
                nodes.push(lowerLevel);
            }, this);
            return nodes;
        } else {
            return node;
        }
    },
    
    bindDataTreeToControlledSeries: function(trees) {
        var reportDefinition = this.dataSet.getReportDefinition();
        var dataSeries = [];
        trees.forEach(function(tree, index) {
            dataSeries.push(this.createStructure({
                type:       "metric",
                name:       reportDefinition.getMetricByIndex(index).getTitle(),
                data:       this.traverseForDataSeries(tree, 0),
                section:    reportDefinition.getMetricGroupPosition().section
            }));
            
        }, this);
        return dataSeries;
    },
    
    traverseForDataSeries: function(node, level, options) {
        var nodes = [];
        // GDC.util.log("length",YAHOO.lang.isArray(node));
        if (YAHOO.lang.isArray(node)) {
            node.forEach(function(child) {
                nodes.push(this.traverseForDataSeries(child, level+1));
            }, this);
            if (!this.isOnTheLowestLevel(level)) {
                GDC.util.log("data", nodes, "are being controlled by level", level, "and series", this.attributeSeries[level]);
                var series = new GDC.charting.data.Series(nodes);
                var cSeries = new GDC.charting.data.ControlledSeries(series, this.attributeSeries[level].data);
                var iSeries = new GDC.charting.data.IteratingSeries(cSeries);
                return iSeries;
            // } else if (nodes.length !== 1){
                // GDC.util.log("On the lowest level", nodes);
                // return new GDC.charting.data.Series(nodes);
            } else {
                // GDC.util.log("NUMBER OF ROW ATTRIBUTES ", this.attributesPerSection.rows.length, "number of nodes", nodes.length);
                if (nodes.length === 1 && this.attributesPerSection.rows.length === 0){
                    // GDC.util.log("On the lowest level", nodes);
                    // return new GDC.charting.data.Series(nodes);
                    GDC.util.log("On the lowest level", nodes, "just one value");
                    return nodes[0];
                } else {
                    return new GDC.charting.data.Series(nodes);
                }
                
            }
        } else {
            return node;
        }
        
    },
    
    isOnTheLowestLevel: function(currentLevel) {
        // GDC.util.log("CURRENT level", currentLevel);
        // GDC.util.log("lowest level ", this.attributeSeries.length - 1);
        return currentLevel > this.attributeSeries.length - 2;
    },
    
    traverseTreeDF: function(node, level, toDo, scope, wantedLevel) {
        var nodes = [];
        if (wantedLevel === level) {
            return toDo.call(scope, node, level);
        };
        if (node.children.length > 0) {
            node.children.forEach(function(child) {
                nodes.push(this.traverseTreeDF(child, level+1, toDo, scope, wantedLevel));
            }, this);
            return new GDC.charting.data.Series(nodes);
        } else {
            return toDo.call(scope, node, level);
        }
    }
    
    
});

GDC.charting.data.TreeNode = function(options) {
    this.id = options.id;
    this.type = options.type;
    this.label = options.label;
    this.distance = options.distance;
};

YAHOO.lang.augmentObject(GDC.charting.data.TreeNode.prototype, {
    
    getLabel: function() {
        return this.label;
    },
    
    getId: function() {
        return this.id;
    },
    
    getType: function() {
        return this.type;
    },
    
    getDistance: function() {
        return this.distance;
    }
});

GDC.charting.data.DataBag = function(attributes, data) {
    this.columnAttributes = attributes.columns;
    this.rowAttributes = attributes.rows;
    this.dataSeries = data;
};

YAHOO.lang.augmentObject(GDC.charting.data.DataBag.prototype, {
    getRowAttributes: function(level) {
        return YAHOO.lang.isNumber(level) ? this.rowAttributes[level] : this.rowAttributes;
    },
    getColumnAttributes: function(level) {
        return YAHOO.lang.isNumber(level) ? this.columnAttributes[level] : this.columnAttributes;
    },
    getMetricSeries: function(level) {
        return YAHOO.lang.isNumber(level) ? this.dataSeries[level] : this.dataSeries;
    }
});

YAHOO.lang.augment(GDC.charting.data.MDataSet, YAHOO.util.EventProvider);