(function(){
    
    function Axis(ggfx, a, availPixels) {
        var tsize = a.style.size;

        // will be calculated by the calculate() call
        var min = 0, max = 0;
        var pixels = 0;
        var offset = 0;
        var ratio = 0;

        // will be set later on by setMargin()
        var margin = {beg:0, end:0};

        this.map = map;

        function map(v) {
            return offset + ( v - min ) * ratio;
        }

        this.getAdvance = function() {
            return this.range.tick * ratio;
        }

        this.origin = function(v) {
            if (min > 0) return map(min);

            return map(0);
        }

        function calculate(avPixels, margin, tsize) {
            function itCount(it) {
                var cnt = 0;
                while (it.hasNext()) { it.next(); cnt++; }
                return cnt;
            }

            if (a.ticks) {
                a.range = a.range || {};
                a.range.min = 0;
                a.range.max = itCount( a.ticks.getIterator()); // FIXME: series length
            }

            // if includeZero then:
            if (true || a.includeZero) {
                min = a.range.min;
                if (min > 0) min = 0;
                max = a.range.max;
                if (max < 0) max = 0;
            }

            // offset for mapping oprations
            pixels = avPixels - margin.beg - margin.end;
            offset = margin.beg;

            // FIXME: ATM we count on having a half of the text size both down
            //        and up overhang in the axis space caused by horizontal tick
            //        labels (FUTURE: add rotated text options etc.)
            pixels -= tsize;
            offset += tsize/2;

            //console.log("%o %d... %d", margin, offset, tsize);

            // This is the actual axis range (the min/max metric values shown
            // on the axis). Normally one would want at least the closest
            // greater than axis range ticks to be visible.
            //
            // FIXME: this should only work for vertical axes
            var axisRange = getTickByFontSize( min, max, pixels, tsize);

            // if this is an attribute axis -> do not extend beyond the
            // designated min/max range
            if (a.ticks) {
                axisRange.min = min;
                axisRange.max = max;
                // number of ticks we can fit in
                axisRange.tick = Math.ceil((max-min) / (pixels / (tsize*2)));
                if (axisRange.tick < 1)
                    axisRange.tick = 1;
            }

            // hmm, margins??
            ratio = pixels / (axisRange.max - axisRange.min);

            min = axisRange.min;
            max = axisRange.max;

            this.range = axisRange;

            // generate label texts
            this.labels = makeLabels( axisRange );

            // alternate line for horizontal labels
            this.labels.alternate = this.labels.maxWidth > this.getAdvance() ? 1 : 0;
        }

        function getTickByFontSize( min, max, availablePixels, fontSize ) {
            function digits(x) {
                return x !== 0 ? Math.floor(Math.log(Math.abs(x))/Math.LN10) : 1;
            }

            // number of digits in the metric tick labels
            var digs = digits(max);
            if ( min !== 0 ) digs = Math.min(digits(min), digs);
            var dim = Math.pow(10, digs);

            // tick distance (number of ticks) determined based
            // on the font size
            var dtick = 0;
            do {
                dim /= 10; // tick label rounding by magnitudes
                dtick = Math.floor(
                        Math.round(max-min) / (availablePixels/fontSize/3.4)
                        / dim) * dim;
            } while (dtick <= 0);

            var dmax = Math.floor(((max+dtick)) / dtick) * dtick;
            var dmin = Math.ceil(((min-dtick)) / dtick) * dtick - dtick/2;

            // if one of the local extremes is 0 then let the last
            // tick be 0 as well (no negative ticks for positive only ranges)
            if (min === 0) dmin = 0;
            if (max === 0) dmax = 0;

            return {min: dmin, max:dmax, tick:dtick};
        }

        function makeLabels( ar ) {
            var ticktexts = [];
            var maxew = -1;

            var t, e;

            if (a.ticks) {
                var it = a.ticks.getIterator();
                var idx = 0;

                //for(var idx=ar.min; idx<=ar.max; idx+=ar.tick) 
                while(it.hasNext() ) {
                    t = it.next();

                    if (idx !== 0) {
                        var tickCnt = ar.tick;
                        while(--tickCnt > 0 && it.hasNext() ) {
                            t = it.next();
                            idx++;
                        }
                    }

                    e = ggfx.getTextExtent(t, a.style.size);

                    // count it into the maximum axis label width
                    maxew = Math.max(maxew, e.w);

                    ticktexts.push({idx:idx, extent:e, text:t});
                    idx++;
                }

                // push the past-end tick (with no text) to allow
                // 'center' alignment
                ticktexts.push({idx:idx});

                return {texts: ticktexts, maxWidth: maxew};
            }

            // 0 ... up
            for(idx=0; idx<=ar.max; idx+=ar.tick) {
                t = String( Math.round(idx * 1000) / 1000 );
                e = ggfx.getTextExtent(t, a.style.size);

                // count it into the maximum axis label width
                maxew = Math.max(maxew, e.w);

                ticktexts.push({idx:idx, extent:e, text:t});
            }

            // 0 ... down
            for(idx=-ar.tick; idx>=ar.min; idx-=ar.tick) {
                t = String( Math.round(idx * 1000) / 1000 );
                e = ggfx.getTextExtent(t, a.style.size);

                // count it into the maximum axis label width
                maxew = Math.max(maxew, e.w);

                ticktexts.push({idx:idx, extent:e, text:t});
            }

            return {texts: ticktexts, maxWidth: maxew};
        }

        calculate.call(this, availPixels, margin, tsize );

        this.getWidth = function() {
            if (a.type === 'y') {
                // get tick label max width + longest tick line lenght
                var w = this.labels.maxWidth + tsize/2 + tsize/4;

                // title as a rotated text (width ~ font size)
                if (a.title) {
                    w += tsize + tsize/4;
                }

                return w;
            }
            return availPixels;
        }
        this.getHeight = function() {
            if (a.type === 'x') {
                var ts = tsize;
                var h = /*font descent*/ ts/4 + /*ticklen*/ ts/4;

                // basically label texts height
                h += ts;

                if (a.title) {
                    h += ts;
                }

                var rot = Math.abs(a.style.rotate);
                if (rot) {
                    var sinRot = Math.sin((rot / 180) * Math.PI);
                    h += this.labels.maxWidth * sinRot;
                } else if (this.labels.alternate) {
                    h += ts;
                }

                return h;

            }
            return availPixels;
        }

        function recalculate() {
            calculate.call(this, availPixels, margin, tsize );
        }

        this.setMargin = function(m) {
            calculate.call(this, availPixels, m, tsize );

            margin = m;
        }

        this.render = function(cc, w, h) {
            var gfx = GDC.charting.gfx;
            var tr = GDC.charting.tr;

            var g = cc.group = new gfx.Group(cc).render();

            var gridStyle = {color:'#bbbbbb', width:1};
            var sclr = this;

            // this rebuilds the labels for the attribute combination 'makeLabels'
            recalculate.call(this);

            var advance = sclr.getAdvance();
            var ticklen = a.style.size/4;
            var tsize = a.style.size;

            if (a.type === 'x') {
                renderXAxis(cc, a);
            } else {
                renderYAxis(cc, a);
            }

            return g;


            function renderXAxis(cc, a) {
                var sx, e, text;
                var sy = -tsize/4; // font descent

                if (a.title) {
                    text = new gfx.Text(g, null, a.title, a.style.size).render();
                    e = text.getExtent();
                    var xtit = sclr.map(sclr.range.min) +
                        (sclr.map(sclr.range.max) - sclr.map(sclr.range.min) - e.w)/2;
                    tr.move(text, xtit, sy); // FIXME shift by Y axis width
                    text.setFill(a.style.fill).setStroke({color:a.style.fill, width:0.5});

                    sy -= tsize;
                }

                // shift some more if alternating lines
                sy -= sclr.labels.alternate * tsize;

                // axis line above its labels
                sy -= tsize;

                // axis line
                if (!a.ticks) {
                    new gfx.Path(cc).render()
                        .moveTo(0, sy-ticklen).lineTo(w, sy-ticklen)
                        .setStroke({color:a.style.fill, width:2.5});
                }

                var line = sclr.labels.alternate;

                // right align the tick labels and set their stroke&fill
                sclr.labels.texts.forEach(function(t) {
                    var gg, text;
                    var e = t.extent;

                    sx = sclr.map(t.idx);
                    var s = { x: sx, y: sy };

                    // gridline
                    if (a.gridlines) {
                        new gfx.Path(cc).render()
                            .setStroke(gridStyle)
                            .moveTo(s.x, 0).lineTo(s.x, s.y);
                    }

                    // tick
                    if (!a.ticks) {
                        new gfx.Path(cc).render()
                            .moveTo(s.x, s.y).lineTo(s.x, s.y-ticklen)
                            .setStroke({color:a.style.fill, width:2.5});
                    }

                    if (!t.text) return;

                    // tick label
                    if (a.style.align === "center") {
                        s.x += advance / 2;
                    }

                    gg = new gfx.Group(cc).render();
                    text = new gfx.Text(gg, null, t.text, e.h).render();

                    if (a.style.rotate) {
                        tr.rotate(text, a.style.rotate);

                        var rot = Math.abs(a.style.rotate);
                        var sinRot = Math.sin((rot / 180) * Math.PI);
                        var cosRot = Math.cos((rot / 180) * Math.PI);
                        if (a.style.rotate > 0) {
                            s.y -= sclr.labels.maxWidth * sinRot + e.h/3;
                            s.x -= e.h/3; // FIXME should probably not be adjusted
                        } else {
                            s.y -= (sclr.labels.maxWidth - e.w) * sinRot + e.h/3;
                            s.x -= e.w * cosRot - e.h/3;
                        }
                    } else {
                        s.x -= e.w/2; // align center wrt the tick

                        line ^= sclr.labels.alternate;

                        //TODO: hide labels that don't fit
                        //if ( e.w > advance )
                        //    text.getNode().style.visibility = 'hidden';
                    }

                    s.y += -ticklen/2 + tsize + tsize * line;
                    tr.move(gg, s.x, s.y);
                    text.setFill(a.style.fill).setStroke({color:a.style.fill, width:0.5});
                });
            }
        
            function renderYAxis(cc, a) {
                sx = 0;

                if (a.title) {
                    text = new gfx.Text(g, null, a.title, tsize).render();
                    e = text.getExtent();
                    var ytit = sclr.map(sclr.range.min) +
                        (sclr.map(sclr.range.max) - sclr.map(sclr.range.min) - e.w)/2;
                    tr.move(text, tsize, h - ytit);
                    tr.rotate(text, -90);
                    text.setFill(a.style.fill).setStroke({color:a.style.fill, width:0.5});

                    sx += tsize + tsize/4;
                }

                sx += sclr.labels.maxWidth + tsize/2;

                // axis line
                if (!a.ticks) {
                    new gfx.Path(cc).render()
                        .moveTo(sx+ticklen, 0).lineTo(sx+ticklen, h - sclr.map(sclr.range.min) + advance/3)
                        .setStroke({color:a.style.fill, width:2.5});
                }

                // right align the tick labels and set their stroke&fill
                sclr.labels.texts.forEach(function(t) {
                    var gg, text;
                    var e = t.extent;

                    sy = h - sclr.map(t.idx);
                    var s = { x: sx, y: sy };

                    // gridline
                    if (a.gridlines) {
                        new gfx.Path(cc).render()
                            .setStroke(gridStyle)
                            .moveTo(s.x, s.y).lineTo(w, s.y);
                    }

                    // tick
                    if (!a.ticks) {
                        new gfx.Path(cc).render()
                            .moveTo(s.x, s.y).lineTo(s.x+ticklen, s.y)
                            .setStroke({color:a.style.fill, width:2.5});
                    }

                    if (!t.text) return;

                    // tick label
                    gg = new gfx.Group(cc).render();
                    text = new gfx.Text(gg, null, t.text, e.h).render();

                    s.x -= e.h/3 + e.w; // align right
                    s.y += e.h/4; // text baseline adjustments

                    if (a.style.align === "center") {
                        s.y -= advance/2;
                    }
                    if (a.style.rotate) {
                        // for left-side axis (labels to the left of ticks)
                        // rotate the text around its lower-right corner
                        tr.move(text, e.w, 0);
                        tr.rotate(text, a.style.rotate);
                        tr.move(text, -e.w, 0);
                    }
                    tr.move(gg, s.x, s.y); //text.setShape(s)

                    text.setFill(a.style.fill).setStroke({color:a.style.fill, width:0.5});
                });
            }
        }
    }

    GDC = GDC || {};
    GDC.charting = GDC.charting || {};

    GDC.charting.Axis = Axis;
})();

