/**
**/

(function(){

 var charting = new function () {

    var MapIterator = GDC.charting.data.MapIterator;
    var ConstSeries = GDC.charting.data.ConstSeries;
    var InspectedSeries = GDC.charting.data.InspectedSeries;
    var SeriesProperties = GDC.charting.data.SeriesProperties;
    var XYZSpecs = GDC.charting.data.XYZSpecs;
    var charting = this;
    var tr = GDC.charting.tr;
    var gfx = GDC.charting.gfx;

    /**
     * Hierarchical context class.
     * @param parent Parent context object
     * @return class
     */
    function Ctx(parent) {
        this.parent = parent;
        this.curr = this;

        // context global object (not hierarchical globals)
        this.global = parent ? parent.global : null;

        this.push = function() {
            this.curr = new Ctx(this.curr);
            return this.curr;
        }

        this.pop = function() {
            return this.curr = this.curr.parent;
        }

        /**
         * Put value into the current context. Similar to what one
         * can do with ctx.push().name = value.
         **/
        this.put = function(name, value) {
            this.curr[name] = value;
        }
        this.get = function(name) {
            var cc = this.curr;
            while (cc) {
                var cv = cc[name];
                if (cv != undefined) return cc;
                cc = cc.parent;
            }
            return cc;
        }
    }

    function getAt(it, n) {
        if (! it.hasNext()) return null;
        
        var res;
        do {
            res = it.next();
        } while(n-- > 0 && it.hasNext());

        return res;
    }

    function showTooltip(e, cc) {
        if (!cc.global.tooltip) return;

        var p = {x: e.pageX, y: e.pageY};

        var bb = cc.global.gfx.getBBox(cc.gfx);
        p = {x: bb[0].x + (bb[2].x - bb[0].x)/2,
             y: bb[0].y + (bb[2].y - bb[0].y)/2};

        /*
        var tt = document.getElementById('tooltip');
        // check whether the id=tooltip element exists
        if (!tt) return;

        tt.style.position = 'absolute';
        tt.style.top = p.y + 'px';
        tt.style.left = p.x + 'px';
        tt.style.color = 'teal';
        tt.style.padding = '6px';
        tt.style.paddingBottom = '4px';
        //Let the HTML renderer calculate the text widths and heights
        //tt.style.width = '40px';
        //tt.style.height = '30px';
        tt.style.zIndex = '1001';
        tt.style.border = 'thin dotted';
        tt.style.borderColor = 'black';
        tt.style.backgroundColor = 'lightyellow';
        */

        // get lookup indexes
        var cci = cc;
        var idxs = [];
        while (cci) {
            var ids = String(cci.id).split(':').reverse();
            ids.forEach(function(id) {
                var idn = id - 0; // convert to number
                if (idn == id) { // really the same number when converted?
                    idxs.unshift( idn );
                }
            });
            cci = cci.parent;
        }

        // construct the tooltip text lines
        var lines = [];
        // lines.push(getId(cc) + "... " + idxs);

        for(var ci = 0; ci < idxs.length; ci++) {
            var idx = idxs[ci];
            if (typeof(idx) !== 'number') continue;

            var d = cc.global.tooltip.content[ci];
            if (!d || !d.data) continue;

            for(var ni = 0; ni < d.data.length; ni++) {
                var line = '';

                // iterate over the data lookup (to influence the
                // dependent iterators)
                var data = (d.data[ni]) ? getAt( d.data[ni].getIterator(), idx) : "";

                var name = d.names[ni];
                if (name === null || !d.data[ni] ) continue;

                line += name.length ? name + ": " : "";

                if (d.lookup && d.lookup[ni]) {
                    line += getAt(d.lookup[ni].getIterator(), idx) + " ; ";
                }

                line += data;
                lines.push( line);
            }
        }

        //tt.innerHTML = getId(cc) + '<br><br>' + text;
        //tt.innerHTML = lines.join("<br>");

        // render the tooltip

        if (! lines.length) {
            // do nothing when there is no text to show
            return;
        }

        var ttip = cc.global.tooltip.style;
        var ttext;

        if (!cc.global.gfxTooltip.children.length) {
            cc.global.gfxTooltip.createRect({x: 0, y: 0, width: 1, height: 1, r:4})
                .setStroke({color: 'gray', width: 2}).setFill('lightyellow');

            ttext = new gfx.MultiText(cc.global.gfxTooltip,
                                lines, ttip.size).render(null).setFill(ttip.fill).setStroke({color:ttip.fill});
        } else {
            ttext = new gfx.MultiText(cc.global.gfxTooltip,
                                lines, ttip.size).render(cc.global.gfxTooltip.children[1]).setFill(ttip.fill).setStroke({color:ttip.fill});

        }

        ttext.setShape({text: lines});
        var extent = ttext.getExtent();

        // set tooltip rect position 
        cc.global.gfxTooltip.children[0].setShape({x: p.x, y: p.y, width: extent.w + ttip.size, height: extent.h + ttip.size*.9});
        ttext.setTransform(dojox.gfx.matrix.translate(p.x + ttip.size/2, p.y + ttip.size/3));

        //console.log("%s %o", lines.join("\n"), extent);
    }

    function fireevent(e) {
        //console.log("event: %s: %o", e.target.cc.gfx.rawNode.innerHTML, e);

        // look for the nearest parent containing the CHOM pointer
        // -> the one of which the 'e.target' is a rendering of
        var t = e.target;
        while(t && !t.cc) {
            t = t.parentNode;
        }
        if (t) {
            // console.log("event: %s: %o", t.cc.gfx.rawNode.innerHTML, e);
            showTooltip(e, t.cc);
        }
    }

    // tooltip hack?
    this.fireevent = fireevent;

    function setStrokeAndFill(cc, ch) {
        cc.s = ( ch.style != undefined ) ? ch.style : cc.style; 
        cc.f = ( ch.style != undefined && ch.style.fill != undefined)
            ? ch.style.fill
            : (cc.style != undefined ? cc.style.fill : undefined);
    }

    function getMargin(cc, t, opts) {
        opts = opts || {};
        opts.style = opts.style || {size: 15, fill:'black'};

        var e = cc.global.gfx.getTextExtent(cc.text, opts.style.size);
        var margin = gfx.getLabelMetrics(cc, opts, e).margin;
        return margin;
    }

    function morphLinearly(curr, from, phase) {
        return from + (curr - from) * phase / 100;
    }

    function morphRect(curr, from, phase) {
        // simple linear coordinate approximation
        curr.x = from.x + (curr.x - from.x) * phase / 100;
        curr.y = from.y + (curr.y - from.y) * phase / 100;
        curr.w = from.w + (curr.w - from.w) * phase / 100;
        curr.h = from.h + (curr.h - from.h) * phase / 100;
    }

    function Ellipse(chart) {
        function xyzStats(cc) {
            // FIXME: what's the minimal size of this bar?
            var s = new XYZSpecs();
            s.x.count = 1;
            s.x.min = 5;
            // Y axis is the same
            s.y = s.x;
            return s;
        }

        function morph(cc, from, phase) {
            morphRect(cc, from, phase);

            /*
            var x = false;
            if (from.h > cc.h && cc.h == 0 && phase > 30) {
                console.log("\\bar: %d, %d [%d]", from.h, cc.h, phase);
                x = true;
            }
            if (x) {
                console.log("/bar: %d", cc.h);
                //cc.parent.xxx = true;
            }
            */
            if (!cc.gfx) {
                console.error('!bar.gfx: %o', cc);
                return;
            }
            update(cc);
        }

        function calculate(cc) {
            setStrokeAndFill(cc, chart);

            cc.text = cc.text || '';
            cc.margin = getMargin(cc, cc.text, chart.labels);
        }

        function update(cc) {
            function renderEllipse(cc, g) {
                return new gfx.Ellipse(cc, g, cc.w/2, cc.h/2).render();
            }

            //cc.gfx = new gfx.Ellipse(cc, cc.gfx, cc.w/2, cc.h/2).render();
            cc.gfx = new gfx.Labeled(cc, cc.gfx,
                    renderEllipse, cc.w, cc.h, cc.text, chart.labels).render();

            gfx.setStyle(cc);
        }


        function render(cc) {
            update(cc);

            cc.gfx.connect('onmousemove', charting, 'fireevent');
            cc.gfx.getNode().cc = cc;
            return cc.gfx;
        }

        this.morph = morph;
        this.calculate = calculate;
        this.xyzStats = xyzStats;
        this.render = render;
    }

    function Bar(chart) {
        function xyzStats(cc) {
            // FIXME: what's the minimal size of this bar?
            var s = new XYZSpecs();
            s.x.count = 1;
            s.x.min = 5;
            // Y axis is the same
            s.y = s.x;

            // calculate the necessary space around the marker
            // (to fit in labels etc.)
            s.margin = cc.margin;

            cc.xyz = s;
            return s;
        }

        function morph(cc, from, phase) {
            morphRect(cc, from, phase);

            if (!cc.gfx) {
                console.error('!bar.gfx: %o', cc);
                return;
            }
            update(cc);
        }

        function calculate(cc) {
            setStrokeAndFill(cc, chart);

            // corner rounding
            cc.rr = chart.style ? chart.style.r : undefined;

            cc.text = cc.text ? String(cc.text) : '';
            // ??? generate XUL-like boxed layout to easily provide XYZ specs
            // do I need it really?
            cc.margin = getMargin(cc, cc.text, chart.labels);
        }

        function update(cc) {
            function renderBar(cc, g) {
                return new gfx.Bar(cc, g, cc.w, cc.h, cc.rr).render();
            }

            cc.gfx = new gfx.Labeled(cc, cc.gfx,
                    renderBar, cc.w, cc.h, cc.text, chart.labels).render();

            //cc.gfx = new gfx.BarText(cc, cc.gfx,
            //                         cc.w, cc.h,
            //                         cc.text ? String(cc.text) : null).render();
            //cc.gfx = new gfx.Bar(cc, cc.gfx, cc.w, cc.h).render();
            gfx.setStyle(cc);
        }


        function render(cc) {
            update(cc);

            cc.gfx.connect('onmousemove', charting, 'fireevent');
            cc.gfx.getNode().cc = cc;
            return cc.gfx;
        }

        this.morph = morph;
        this.calculate = calculate;
        this.xyzStats = xyzStats;
        this.render = render;
        this.renderAxes = function(){ return null; };
    }

    function PieSlice(chart) {
        function calculate(cc) {
            setStrokeAndFill(cc, chart);
    
            cc.margin = getMargin(cc, cc.text, chart.labels);
        }

        function xyzStats(cc) {
            // dummy - only used in Pie() which doesn't iterate
            // over it children
            return new XYZSpecs();
        }

        function morph(cc, from, phase) {
            cc.a = from.a + (cc.a - from.a) * phase / 100;
            cc.rotation = from.rotation + (cc.rotation - from.rotation) * phase / 100;
            cc.w = from.w + (cc.w - from.w) * phase / 100;
            cc.h = from.h + (cc.h - from.h) * phase / 100;

            update(cc);
        }

        function update(cc) {
            function renderSlice(cc, g) {
                return new gfx.PieSlice(cc, g, cc.a, cc.w/2).render();
            }

            cc.gfx = new gfx.Labeled(cc, cc.gfx,
                    renderSlice, cc.w, cc.h, cc.text, chart.labels).render();

            gfx.setStyle(cc);

            // FIXME: ugly knowing that the pieslice is the first child
            // in Labeled
            cc.gfx.children[0].setTransform(
                    dojox.gfx.matrix.rotateg(cc.rotation));
        }

        function render(cc) {
            update(cc);

            cc.gfx.connect('onmousemove', charting, 'fireevent');
            cc.gfx.getNode().cc = cc;
            return cc.gfx;
        }

        this.morph = morph;
        this.calculate = calculate;
        this.xyzStats = xyzStats;
        this.render = render;
    }


    function Pie(chart) {
        function calculate(cc) {
            var ch = chart;
            cc.children = [];

            setStrokeAndFill(cc, chart);
            cc.margin = getMargin(cc, cc.text, chart.labels);

            var a = ch.series['width'];
            var propsA = new InspectedSeries(a).props;
            var da = 360/propsA.sum;

            var r = ch.series['radius'];
            if (r === undefined) {
                r = new InspectedSeries(
                            new ConstSeries(1, 1));
            }
            var propsR = new InspectedSeries(r).props;
            var dr = cc.w/propsR.max;

            var colors = ch.series['color'];
            var texts = ch.series['text'];

            var a_it = a.getIterator();
            var r_it = r.getIterator();
            var colors_it = colors.getIterator();
            var texts_it = (texts != undefined) ? texts.getIterator() : null;

            var rotation = 0;
            var n = 0;
            while(a_it.hasNext()) {
                var e = cc.push();
                e.id = n++; // just an element number
                e.what = ch.what;
                cc.pop();
                cc.children.push(e);

                e.a = a_it.next() * da;
                e.rotation = rotation;
                e.w = e.h = r_it.next() * dr;
                // HACK: fill color only? what about generic style?
                e.style = {fill:colors_it.next()};
                if (texts_it) e.text = texts_it.next();

                e.what.handler.calculate(e);
                rotation += e.a;
            }
        }

        function xyzStats(cc) {
            var s = new XYZSpecs();
            s.x.count = 1;
            s.x.min = 20; // min Pie size
            // Y axis is the same
            s.y = s.x;
            return s;
        }

        function morph(cc, from, phase) {
            morphRect(cc, from, phase);

            update(cc);
        }

        function renderGfx(cc, gg) {
            if (gg) return gg;

            var pie = cc.group = new gfx.Group(cc).render();

            // create the pie origin shift group
            var g = cc.group = new gfx.Group(cc).render();

            cc.children.forEach(function(e) {
                e.group = g;
                e.what.handler.render(e);
            });

            return pie;
        }

        function update(cc) {
            // no need to update the children, PieSlice children
            // are updated just like any other leaf shapes
            cc.gfx = new gfx.Labeled(cc, cc.gfx,
                    renderGfx, cc.w, cc.h, cc.text, chart.labels).render();
            gfx.setStyle(cc);

            // FIXME: ugly knowing the internals of cc.gfx at all!!

            // just move the origin of the children PieSlices
            cc.gfx.children[0].children[0].setTransform(
                dojox.gfx.matrix.translate(cc.w/2,cc.h/2));
        }

        function render(cc) {
            update(cc);

            return cc.gfx;
        }

        this.morph = morph;
        this.calculate = calculate;
        this.xyzStats = xyzStats;
        this.render = render;
    }


    function Overlay(chart) {
        function prepare(chart) {
            var it = chart.series.getIterator();
            // recurse over all children
            while(it.hasNext()) {
                prepareHandlers(it.next());
            }
        }

        function calculate(cc) {
            var ch = chart;
            cc.children = [];

            calculateAxes(cc, ch);

            // recurse
            var d = chart.series.getIterator();
            var n = 0;
            while (d.hasNext()) {
                // overlay does NOT push any context
                var e = cc.push();
                cc.pop();
                cc.children.push(e);

                e.id = n++;
                e.style = cc.style;
                e.xaxis = cc.xaxis;
                e.yaxis = cc.yaxis;
                e.w = cc.w;
                e.h = cc.h;
                e.what = d.next();
                e.what.handler.calculate(e);
            }
        }

        function xyzStats(cc) {
            // recurse
            var s = new XYZSpecs();

            var d = chart.series.getIterator();
            while (d.hasNext()) {
                var sd = d.next().handler.xyzStats(cc);
                sd.x.count = Math.max(s.x.count, sd.x.count);
                sd.y.count = Math.max(s.y.count, sd.y.count);
                s.x.count = 0;
                s.y.count = 0;
                s.merge(sd);
            }

            return s;
        }

        function render(cc) {
            var oldg = cc.group;
            var g = cc.group = new gfx.Group(cc).render();
            cc.gfx = g;
            cc.children.forEach(function(e) {
                e.group = cc.group;
                e.what.handler.render(e);
                cc.group = g;
            });
            cc.group = oldg;
            return g;
        }

        this.prepare = prepare;
        this.calculate = calculate;
        this.xyzStats = xyzStats;
        this.renderAxes = renderCommonAxes;
        this.render = render;
    }


    function Marker(chart) {
        function calculate(cc) {
            if (typeof cc.marker === 'string') switch(cc.marker) {
            case 'ELLIPSE': 
                cc.marker = new Ellipse(chart);
                break;
            case 'BAR': 
                cc.marker = new Bar(chart);
                break;
            }
            
            cc.marker.calculate(cc);
        }

        function xyzStats(cc) {
            return cc.marker.xyzStats(cc);
        }

        function morph(cc, from, phase) {
            return cc.marker.morph(cc, from, phase);
        }

        function render(cc) {
            return cc.marker.render(cc);
        }

        this.morph = morph;
        this.calculate = calculate;
        this.xyzStats = xyzStats;
        this.render = render;
    }

    function Scattered(chart, connectMarkers) {
        function calculate(cc) {
            var ch = chart;
            cc.children = [];

            cc.style = ch.style;

            var x = ch.series['x'];
            var y = ch.series['y'];
            var sizes = ch.series['size'];
            var colors = ch.series['color'];
            var texts = ch.series['text'];
            var markers = ch.series['marker'];
            if (markers === undefined) {
                markers = new ConstSeries(ch.what.type);
            }

            var ds;
            if (sizes === undefined) {
                sizes = new InspectedSeries(new ConstSeries(1, 1));
                ds = cc.style.width*3;
                var propsS = {min:ds, max:ds};
            } else {
                var propsS = sizes.props;
                ds = Math.min(cc.h,cc.w)/((propsS.max-propsS.min)*propsS.count*2);
            }
            
            var rmin = propsS.max * ds / 4;
            calculateAxes(cc, ch, { x: {beg:rmin, end:rmin},
                                    y: {beg:rmin, end:rmin} });

            var x_it = x.getIterator();
            var y_it = y.getIterator();
            var sizes_it = sizes.getIterator();
            var colors_it = colors.getIterator();
            var texts_it = (texts != undefined) ? texts.getIterator() : null;
            var markers_it = markers.getIterator();

            var n = 0;
            while(x_it.hasNext() && y_it.hasNext()) {
                var px = cc.xaxis.map( x_it.next() );
                var py = cc.h - cc.yaxis.map( y_it.next() );
                var ps = sizes_it.next() * ds;

                // console.log("x: %d, y: %d", x_it.peek(), y_it.peek());
                var e = cc.push();
                e.id = n++; // marker number
                e.what = ch.what;
                cc.pop();
                cc.children.push(e);

                e.w = ps;
                e.h = ps;
                // HACK: fill color only? what about generic style?
                e.style = {fill:colors_it.next()};
                e.marker = markers_it.next();
                if (texts_it) e.text = texts_it.next();

                if (x_it.peek() === null || y_it.peek() === null) {
                    e.w = 0;
                    e.h = 0;
                }

                e.x = px-e.w/2;
                e.y = py-e.h/2;
                e.what.handler.calculate(e);
            }
        }

        function update(cc) {
            var e;

            if (connectMarkers) {
                var path = cc.gfx.children[0];
                path.setShape("");

                cc.children.forEach(function(e) {
                    // FIXME: TODO, missing values? what to do here?
                    if (true || e.w !== 0 && e.h !== 0) {
                        var x = e.x+e.w/2;
                        var y = e.y+e.h/2;
                        if (path.getShape().path.length > 0)
                            path.lineTo(x,y);
                        else
                            path.moveTo(x,y);
                    }

                    // set the marker position
                    e.gfx.setTransform(dojox.gfx.matrix.translate(e.x,e.y));
                });
            } else {
                cc.children.forEach(function(e) {
                    // set the marker position
                    e.gfx.setTransform(dojox.gfx.matrix.translate(e.x,e.y));
                });
            }
        }

        function morph(cc, from, phase) {
            // scattered morphing is done by morphing its children sizes
            // and updating the marker positions and connecting line
            // if necessary
            update(cc);
        }

        function render(cc) {
            var ch = chart;
            var g = cc.group = new gfx.Group(cc).render();
            cc.gfx = g;

            // iterate over the chart values
            if (connectMarkers)
                new gfx.Path(cc).render().setStroke(cc.style);

            cc.children.forEach(function(e) {
                // render the markers
                e.group = g;
                e.what.handler.render(e);
            });

            update(cc);
            return g;
        }

        this.morph = morph;
        this.calculate = calculate;
        this.xyzStats = function(){ return new XYZSpecs(); };
        this.renderAxes = renderCommonAxes;
        this.render = render;
    }

    function Grid(chart) {
        function calculate(cc) {
            var ch = chart;
            cc.children = [];

            calculateAxesGridTree(cc, ch);

            // inspect/iterate over the control series and recurse
            var colsProps = new SeriesProperties();
            var rowsProps = new SeriesProperties();
            var rows = ch.series['rows'];
            var cols = ch.series['cols'];
            var rows_it = rows.getIterator();
            var cols_it;

            while(rows_it.hasNext()) {
                rowsProps.mergeElement( rows_it.next() );

                var cProps = new SeriesProperties();
                cols_it = cols.getIterator();
                while(cols_it.hasNext()) {
                    cProps.mergeElement( cols_it.next() );

                    // recurse
                    var chc = cc.push();
                    cc.pop();
                    cc.children.push(chc);
                }
                colsProps.merge(cProps);
            }

            // Make it available precalculated to the 'render' method
            cc.props = { rows: rowsProps, cols: colsProps };

            // FIXME: THIS needs to be variable based on the xyzStats results
            //var width = cc.w/cc.props.cols.count - ch.pad*2;
            //var height = cc.h/cc.props.rows.count - ch.pad*2;

            var width = cc.xaxis.getAdvance() - ch.pad;
            var height = cc.yaxis.getAdvance() - ch.pad;

            // Calculate the cell sizes and positions
            var sr = new XYZSpecs();

            var cellIdx = 0;
            var row = 1;

            cc.rows = [];

            rows_it = rows.getIterator();
            while(rows_it.hasNext()) {
                rows_it.next();

                var sc = new XYZSpecs();

                var trow = [];

                var col = 0;
                cols_it = cols.getIterator();
                while(cols_it.hasNext()) {
                    cols_it.next();

                    var e = cc.children[cellIdx++];
                    e.id = row-1 +':'+ col; // just an element number
                    e.w = width;
                    e.h = height;
                    e.x = cc.xaxis.map(col) + ch.pad/2;
                    e.y = cc.h - cc.yaxis.map(row) + ch.pad/2;

                    // do not render axes inside grid cells
                    e.xAxes = false;
                    e.yAxes = false;
    
                    e.what = ch.what;
                    e.what.handler.calculate(e);

                    // XYZ
                    var scc = e.what.handler.xyzStats(e);
                    scc.y.count = Math.max(sc.y.count, scc.y.count);
                    sc.y.count = 0;
                    sc.merge( scc );

                    trow.push(e);
                    col++;
                }

                cc.rows.push(trow);

                //xxx = xxx + ", " + sc.x.count + ":" + ch.what.type;
                //xxx = xxx + "\n";
                // sc.y.count /= col_count;
                sc.x.count = Math.max(sr.x.count, sc.x.count);
                sr.x.count = 0;
                sr.merge(sc);

                row++;
            }

            cc.xyz = sr;
        }

        function xyzStats(cc) {
            return cc.xyz;
        }

        function renderAxes(cc, type, rcc) {
            var ch = chart;
            var g = cc.group = new gfx.Group(cc).render();

            var axs = ch.axes;
            if (!axs) {
                return g;
            }

            var a, ca, ra;
            axs.forEach(function(a) {
                if (a.type === 'x') {
                    ca = a;
                } else if (a.type === 'y') {
                    ra = a;
                }
            });

            var xoff = ra.axis.getWidth();
            var yoff = ca.axis.getHeight();

            var oldg, ga;

            if (type === 'y') {
                ga = ra.axis.render(cc, cc.w, cc.h);
                cc.group = g;

                // render the grid's cell Y axes recursively
                var it = ch.series['rows'].getIterator();
                cc.rows.forEach(function(row) {
                    it.next();
                    // get the first cell from each row
                    var re = row[0];

                    console.group("axis Y: %s [%d,%d] %o", re.what.type, xoff, re.y, re);

                    re.group = g;
                    ga = re.what.handler.renderAxes(re, type, rcc);
                    ga.rawNode.id = re.what.type;
                    console.groupEnd();

                    if (ga) tr.move(ga, xoff, re.y);
                });
            }

            if (type === 'x') {
                ga = ca.axis.render(cc, ca.axis.getWidth(), ca.axis.getHeight());
                cc.group = g;

                // render the grid's cell X axes recursively
                var it = ch.series['cols'].getIterator();
                cc.rows[cc.rows.length - 1].forEach(function(ce) {
                    it.next();
                    // ce - a cell in the last row
                    console.group("axis X: %s [%d,%d] %o", ce.what.type, ce.x, yoff, ce);

                    ce.group = g;
                    ga = ce.what.handler.renderAxes(ce, type, rcc);
                    ga.rawNode.id = ce.what.type;
                    console.groupEnd();

                    if (ga) tr.move(ga, ce.x, -yoff);
                });
            }

            return g;
        }

        function render(cc) {
            var ch = chart;
            var g = cc.group = new gfx.Group(cc).render();
            cc.gfx = g;

            var gl, a;
            var rows = ch.series['rows'];
            var cols = ch.series['cols'];

            var cellstyles = ch.series['cell_styles'];

            var widthSeries = new ConstSeries(cc.xaxis.getAdvance() - ch.pad/2);
            var heightSeries = new ConstSeries(cc.yaxis.getAdvance() - ch.pad/2);

            var gridlines = ch.gridlines != undefined
                ? ch.gridlines : false;

            /**
            // Grid's bounding rectangle...
             **/

/*
            new gfx.Path(cc).render().setStroke(gridlines)
            .moveTo(0,0).hLineTo(cc.w).vLineTo(cc.h)
            .hLineTo(0).closePath();
            */

            var cellstylegroup = new gfx.Group(cc).render();

            cc.cellstyles = cellstyles != undefined
                ? cellstyles.getIterator() : null;
            cc.heightOffset = new MapIterator(heightSeries.getIterator(),
                    function(v, pad) { return v+pad; },
                    ch.pad*2 );
            cc.row = rows.getIterator();
            cc.height = heightSeries.getIterator();
            var cellIdx = 0;
            var row = 1;
            while(cc.row.hasNext()) {
                cc.row.next();
                var yoff = cc.heightOffset.next();
                cc.widthOffset = new MapIterator(widthSeries.getIterator(),
                        function(v, pad) { return v+pad; },
                        ch.pad*2 );
                cc.width = widthSeries.getIterator();

                // horizontal grid lines
                if (gridlines && cc.row.hasNext()) {
                    a = cc.xaxis;
                    gl = new gfx.Line(cc, a.map(a.range.max) - a.map(a.range.min), 0).render();
                    tr.move(gl, a.map(a.range.min), cc.h - cc.yaxis.map(row));
                    gl.setStroke(gridlines);
                }

                cc.col = cols.getIterator();
                var col = 0;
                while(cc.col.hasNext()) {
                    cc.col.next();
                    var xoff = cc.widthOffset.next();

                    if (cc.cellstyles) {
                        var ggg = cc.group;
                        cc.group = cellstylegroup;
                        cc.cellstyle = {width:0, fill: cc.cellstyles.next()};

                        // FIXME: HACK need to fix the style setup
                        //        and other stuff
                        var cg = new gfx.Bar(cc, null, xoff, yoff).render();
                        tr.move(cg,    xoff * col, -yoff * row + cc.h);
                        cg.setStroke(cc.cellstyle).setFill(cc.cellstyle.fill);
                        cc.group = ggg;
                    }

                    // vertical grid lines
                    if (gridlines && row === 1 && col > 0) {
                        a = cc.yaxis;
                        gl = new gfx.Line(cc, 0, a.map(a.range.max) - a.map(a.range.min)).render();
                        tr.move(gl, cc.xaxis.map(col), cc.h - a.map(a.range.max));
                        gl.setStroke(gridlines);
                    }

                    var e = cc.children[cellIdx++];
                    e.group = cc.group;

                    tr.move(e.what.handler.render(e), e.x, e.y);
                    col++;
                }
                row++;
            }

            return g;
        }

        this.morph = function(){};
        this.calculate = calculate;
        this.xyzStats = xyzStats;
        this.renderAxes = renderAxes;
        this.render = render;
    }

    function Columns(chart) {
        function calculate(cc) {
            var ch = chart;
            cc.children = [];

            calculateAxes(cc, ch);

            // inspect/iterate over the control series and recurse
            var props = new SeriesProperties();
            var heights = ch.series['height'];
            var heights_it = heights.getIterator();
            while(heights_it.hasNext()) {
                props.mergeElement( heights_it.next() );

                // recurse
                var chc = cc.push();
                cc.pop();
                cc.children.push(chc);
            }

            // make it available precalculated to the 'render' method
            cc.props = { heights: props };

            cc.xyz = new XYZSpecs();

            var colors_it = ch.series['colors'].getIterator();
            var yzero = cc.yaxis.origin();

            //var width = cc.w/cc.props.heights.count - ch.pad;
            var width = cc.xaxis.getAdvance() - ch.pad;
            var halfpad = ch.pad/2;

            var texts = ch.series['text'];

            var texts_it = (texts != undefined) ? texts.getIterator() : null;
            heights_it = heights.getIterator();

            var n = 0;
            while(heights_it.hasNext()) {
                var e = cc.children[n];
                e.id = n; // just an element number
                e.w = width;
                e.h = cc.yaxis.map( heights_it.next() );
                var h = e.h - yzero;
                if (e.h > yzero) {
                    e.y = yzero + h;
                    e.h = h;
                } else {
                    h = yzero - e.h;
                    e.y = yzero;
                    e.h = h;
                }

                e.x = cc.xaxis.map(n) + halfpad;
                e.y = cc.h - e.y;
                e.fill = colors_it.next();
                e.style = {fill: e.fill};
                if (texts_it) e.text = texts_it.next();

                e.what = ch.what;
                e.what.handler.calculate(e);

                var exyz = e.what.handler.xyzStats(e);
                cc.xyz.merge( exyz );

                // FIXME: get left/right margin
                // (left form the first child, right from the last one)
                cc.xyz.margin = cc.xyz.margin || { top: 0, right: 0, bottom: 0, left: 0 };
                cc.xyz.margin.top = Math.max(cc.xyz.margin.top, exyz.margin.top);
                cc.xyz.margin.bottom = Math.max(cc.xyz.margin.bottom, exyz.margin.bottom);
                n++;
            }

            // columns do not contain any leaves in the Y axis
            cc.xyz.y.count = 1;
        }

        function xyzStats(cc) {
            //console.log('xyz %o', cc.xyz);
            return cc.xyz;
        }

        function render(cc) {
            var ch = chart;
            var g = cc.group = new gfx.Group(cc).render();
            cc.gfx = g;

            var heights = ch.series['height'];
            var values = heights!==null ? heights.getIterator() : null;

            // No series for the column 'height' -> empty group
            if (values === null) {
                return g;
            }

            if (heights.props.min < 0) {
                var xoff = cc.xaxis.origin();
                var ypos = cc.h - cc.yaxis.origin();
                var al = new gfx.Path(cc).render()
                    .setStroke({color:'black', width:1});
                al.moveTo(xoff, ypos).lineTo(cc.xaxis.map(cc.xaxis.range.max), ypos);
            }

            // iterate over the chart values
            var n = 0;
            while(values.hasNext()) {
                values.next();

                var e = cc.children[n++];
                e.group = cc.group;

                tr.move(e.what.handler.render(e), e.x, e.y);
            }

            return g;
        }

        function update(cc) {
            // Columns actuall change its Y position in {top-left=0,0}
            // coordinate system
            // -> need to update their positions when morphed
            cc.children.forEach(function(e) {
                // set the marker position
                e.gfx.setTransform(dojox.gfx.matrix.translate(e.x,e.y));
            });
        }

        function morph(cc, from, phase) {
            update(cc);
        }

        this.morph = morph;
        this.calculate = calculate;
        this.xyzStats = xyzStats;
        this.renderAxes = renderCommonAxes;
        this.render = render;
    }

    function Rows(chart) {
        function calculate(cc) {
            var ch = chart;
            setStrokeAndFill(cc, chart); // for now used just for labels

            cc.margin = getMargin(cc, cc.text, chart.labels);

            cc.xyz = new XYZSpecs();

            cc.children = [];

            var e, n = 0;

            // inspect/iterate over the control series and recurse
            var props = new SeriesProperties();
            var texts = ch.series['text'];

            var sizes = ch.series['size'];
            var sizes_it = sizes.getIterator();
            while(sizes_it.hasNext()) {
                props.mergeElement( sizes_it.next() );

                // recurse
                e = cc.push();
                e.id = n++; // just an element number
                e.what = ch.what;
                cc.pop();
                cc.children.push(e);
            }
            sizes.props = props;

            var colors = ch.series['colors'].getIterator();
            var dy = cc.h/sizes.props.sum;
            var yoff = cc.h;
            n = 0;

            var texts_it = (texts != undefined) ? texts.getIterator() : null;
            sizes_it = sizes.getIterator();
            while(sizes_it.hasNext()) {
                e = cc.children[n++];
                e.w = cc.w;
                e.h = sizes_it.next() * dy;
                yoff -= e.h;
                e.x = 0;
                e.y = yoff;
                e.fill = colors.next();
                e.style = {fill: e.fill};
                if (texts_it) e.text = texts_it.next();
                e.what.handler.calculate(e);

                var exyz = e.what.handler.xyzStats(e);
                cc.xyz.merge( exyz );

                // FIXME: get left/right margin
                // (left form the first child, right from the last one)
                cc.xyz.margin = cc.xyz.margin || { top: 0, right: 0, bottom: 0, left: 0 };
                cc.xyz.margin.top = Math.max(cc.xyz.margin.top, exyz.margin.top);
                cc.xyz.margin.bottom = Math.max(cc.xyz.margin.bottom, exyz.margin.bottom);
            }

            // rows do not contain any leaves in the X axis
            cc.xyz.x.count = 1;
            cc.xyz.margin.top += cc.margin.top;
            cc.xyz.margin.bottom += cc.margin.bottom;

            // TODO: CALCULATE the children minimal sizes and set
            // measurements to be used by render()
        }

        function xyzStats(cc) {
            return cc.xyz;
        }

        function renderGfx(cc) {
            var ch = chart;
            var g = cc.group = new gfx.Group(cc).render();
            cc.gfx = g;

            var sizes = ch.series['size'];
            var sizes_it = sizes!==null ? sizes.getIterator() : null;

            // No series for the column 'height' -> empty group
            if (sizes_it === null) {
                return g;
            }

            var n = 0;
            while(sizes_it.hasNext()) {
                sizes_it.next();

                var e = cc.children[n];
                e.group = cc.group;
                tr.move(e.what.handler.render(e), e.x, e.y);
                n++;
            }
            return g;
        }

        function render(cc) {
            // no need to update the children, PieSlice children
            // are updated just like any other leaf shapes
            cc.gfx = new gfx.Labeled(cc, cc.gfx,
                    renderGfx, cc.w, cc.h, cc.text, chart.labels).render();
            gfx.setStyle(cc);

            return cc.gfx;
        }

        this.morph = function(){}; // FIXME: need to set positions
        this.calculate = calculate;
        this.xyzStats = xyzStats;
        this.render = render;
    }

    function prepareHandlers(chart) {
        if (!chart) return null;

        var handler = null;
        switch(chart.type) {
        case 'LABEL': 
            handler = new Label(chart);
            break;
        case 'ELLIPSE': 
            handler = new Ellipse(chart);
            break;
        case 'BAR': 
            handler = new Bar(chart);
            break;
        case 'PIESLICE': 
            handler = new PieSlice(chart);
            break;
        case 'pie': 
            handler = new Pie(chart);
            break;
        case 'scattered': 
            handler = new Scattered(chart, false);
            break;
        case 'line': 
            handler = new Scattered(chart, true);
            break;
        case 'grid': 
            handler = new Grid(chart);
            break;
        case 'columns': 
            handler = new Columns(chart);
            break;
        case 'rows': 
            handler = new Rows(chart);
            break;
        case 'marker': 
            handler = new Marker(chart);
            break;
        case 'overlay': 
            handler = new Overlay(chart);
            break;
        }

        if (handler && handler.prepare) {
            handler.prepare(chart);
        }
        chart.handler = handler;

        // recurse (chart.what is mostly the only child)
        prepareHandlers(chart.what);

        return handler;
    }

    function calculateAxes(cc, chart, border) {
        // if 'overlay' then we should already have axes calculated
        if (cc.xaxis || cc.yaxis)
            return;

        var axes = cc.what.axes;

        // FIXME: TODO unused at this point (e.g. to cope with the
        // marker sizes in Bubble chart, where the bubble sizes
        // could go beyond the chart area -perhaps they even could
        // in fact)
        var b = border || { x: {beg:0, end:0}, y: {beg:0, end:0} };

        var xa = [], ya = [];

        var xHeight = -1;
        var yWidth = -1;

        var a, axis;
        if (axes) axes.forEach(function(a) {
            if (a.type === 'x') {
                a.axis = new GDC.charting.Axis(cc.global.gfx, a, cc.w);
                xa.push(a);

                // IMPORTANT: the X axis height calculation MUST NOT 
                // be dependent on its margin set (otherwise this
                // algorithm fails)
                xHeight = Math.max( xHeight, a.axis.getHeight() );
            } else if (a.type === 'y') {
                a.axis = new GDC.charting.Axis(cc.global.gfx, a, cc.h);
                ya.push(a);
            }
        });

        // render Y axes
        ya.forEach(function(a) {
            if (cc.xAxes)
                a.axis.setMargin({beg: xHeight, end: 0});
            cc.yaxis = a.axis;

            yWidth = Math.max( yWidth, a.axis.getWidth() );
        });

        // render X axes
        xa.forEach(function(a) {
            if (cc.yAxes)
                a.axis.setMargin( {beg: yWidth, end: 0} );
            cc.xaxis = a.axis;
        });
    }

    function calculateAxesGridTree(cc, chart, border) {
        var axes = cc.what.axes;

        // FIXME: TODO unused at this point (e.g. to cope with the
        // marker sizes in Bubble chart, where the bubble sizes
        // could go beyond the chart area -perhaps they even could
        // in fact)
        var b = border || { x: {beg:0, end:0}, y: {beg:0, end:0} };

        var xa = [], ya = [];

        function calcChildren(ggfx, xa, ya, chart, w, h) {
            if (!chart) return;

            var axes = chart.axes;
            var a, xadv, yadv;
            if (axes) axes.forEach(function(a) {
                if (a.type === 'x') {
                    a.axis = new GDC.charting.Axis(ggfx, a, w);
                    xadv = a.axis.getAdvance();

                    xa.push(a);
                } else if (a.type === 'y') {
                    a.axis = new GDC.charting.Axis(ggfx, a, h);
                    yadv = a.axis.getAdvance();

                    ya.push(a);
                }
            });

            calcChildren(ggfx, xa, ya, chart.what, xadv, yadv);
        }

        calcChildren(cc.global.gfx, xa, ya, chart, cc.w, cc.h);

        var xHeight = -1;
        var yWidth = -1;

        // get max height of the X axes
        xa.forEach(function(a) {
            // IMPORTANT: the X axis height calculation MUST NOT 
            // be dependent on its margin set (otherwise this
            // algorithm fails)
        //    xHeight = Math.max( xHeight, a.axis.getHeight() );
            xHeight += a.axis.getHeight();
        });

        // set the Y axis margin (caused by the X axes)
        if (cc.xAxes)
            ya[0].axis.setMargin({beg: xHeight, end: 0});

        // get width of the Y axes
        ya.forEach(function(a) {
            // after the setMargin
            //yWidth = Math.max( yWidth, a.axis.getWidth() );
            yWidth += a.axis.getWidth();
        });


        if (cc.yAxes)
            xa[0].axis.setMargin( {beg: yWidth, end: 0} );

        cc.yaxis = ya[0].axis;
        cc.xaxis = xa[0].axis;
    }

    function renderCommonAxes(cc, type, rcc) {
        // FIXME: chart margins!
        var m = cc.xyz && cc.xyz.margin ? cc.xyz.margin : {top:0, bottom:0};

        var g = cc.group = new gfx.Group(cc).render();
        //cc.group.createShape({type:'rect', width:cc.w, height:cc.h}).setFill([250,200,200,0.3]);

        cc.what.axes.forEach(function(a) {
            if (a.type !== type) return;

            var ga = a.axis.render(cc, a.axis.getWidth(), a.axis.getHeight());
            cc.group = g;
        });
        return g;
    }

    function calculate(ctx, chart, w, h) {
        // set chart.handler recursively
        prepareHandlers(chart);

        var cc = ctx.push();
        cc.id = 'r';
        cc.what = chart;
        cc.w = w;
        cc.h = h;
        cc.xAxes = true;
        cc.yAxes = true;

        // inspect handers and the data series
        cc.what.handler.calculate(cc);
        cc.what.handler.xyzStats(cc);

        ctx.pop();
        return cc;
    }
var _aaa = 0;

    function chart(chart, w, h, surface) {
        var ctx = new Ctx();
        ctx.global = {
            gfx: new gfx.Gfx( surface ),
            w: w,
            h: h,

            tooltip: chart.tooltip
        };

        // get the tree chart representation
        var cc = calculate(ctx, chart, w, h);

        cc._aaa = ++_aaa;
        return cc;
    }

    function axes(cc, group) {
        // go over the children
        var ag, oldg = cc.group;

        cc.group = group;
        ag = cc.what.handler.renderAxes(cc, 'y', cc);
        cc.group = group;
        ag = cc.what.handler.renderAxes(cc, 'x', cc);
        tr.move(ag, 0, cc.h);
        cc.group = oldg;
        return ag;
    }

    function draw(cc, group) {
        // propagate the external rendering element
        cc.group = group;

//console.log("==BEG==");
//console.log(dumpChart(cc));

        // render the chart
        var r = cc.what.handler.render(cc);

        cc.global.gfxTooltip = group.createGroup();

//console.log("==END==");
//console.log(dojox.gfx.utils.toJson(cc.gfx, true));
        return r;
    }

    function dumpChart(cc, indent) {
        indent = indent ? indent : '';
        var s = '[' + cc.id + ']:' + cc.what.type;
        s += ' w = ' + Math.round(cc.w);
        s += ' h = ' + Math.round(cc.h);
        if (!isNaN(cc.x)) s += ' x = ' + Math.round(cc.x);
        if (!isNaN(cc.y)) s += ' y = ' + Math.round(cc.y);

        s = '(' + s;
        cc.children.forEach(function(e) {
            //console.group();
            s += '\n' + dumpChart(e, indent + '  ');
            //console.groupEnd();
        });
        s += ')';
        return indent + s;
    }

    function cloneChartElement(cc, other) {
        cc.id = other.id;
        cc.x = other.x;
        cc.y = other.y;
        cc.w = other.w;
        cc.h = other.h;
        cc.s = other.s;
        cc.f = other.f;
        cc.a = other.a; // for 'pieslice'
        cc.text = other.text; // for marker/composite labeling
        cc.rotation = other.rotation; // for 'pieslice'
        cc.fill = other.fill;
        cc.what = other.what;
        cc.marker = other.marker; // for 'marker'

        cc.children = [];
        if (other.children) other.children.forEach(function(och) {
            var ch = cc.push();
            cc.children.push( cloneChartElement( ch, och ) );
            cc.pop();
        });
        return cc;
    }

    function cloneChart(other) {
        var cc = new Ctx();
        cloneChartElement(cc, other);
        return cc;
    }

    function getId(cc) {
        if (! cc)
            return '';
        return getId(cc.parent) + (cc.id!=undefined ? ';'+cc.id : '');
    }

    function idsMap(cc, map) {
        map = map ? map : {};
        if (!cc)
            return map;

        map[getId(cc)] = cc;

        if (cc.children) cc.children.forEach(function(e) {
            idsMap( e, map );
        });
        return map;
    }

    function morphShape(from, to, phase) {
        if (phase>=100)
            return to;
        if (to.what.type != from.what.type) {
            console.error('!morphShape.what.type: %o -> %o', from, to);
            return to;
        }

        /* FIXME? let's assume the .gfx representation is the same */
        to.group = from.group;
        to.gfx = from.gfx;
        //console.log('morph: %s %o', to.id, to.gfx);

        to.what.handler.morph(to, from, phase);
        return to;
    }

    function morph(from, to, phase) {
        var m = cloneChart(to);

        console.info('morph: %d -> %d [%d]', from._aaa, to._aaa, phase);

        if (phase > 100) {
            console.warn('!morph.phase >= 100: %d', phase);
            return m;
        }
        if (phase === 0) {
            return null; // FIXME: should we return something here at all?
        }

        // try to match objects from destination in the source chart
        var mIdMap = idsMap(m);
        var fromIdMap = idsMap(from);

        var keys = [];
        for(i in mIdMap) {
            if (mIdMap.hasOwnProperty(i)) { keys.push(i); }
        }

        // morph objects with the longest ids first
        keys.sort(function(a,b) {
                    return a.length > b.length ? -1 : 1; }
                 ).forEach(function(mfi) {
            if (mfi.length === 0) return;
        // for(mfi in mIdMap) {

            var f = fromIdMap[mfi];
            var t = mIdMap[mfi];

            //console.log('morph.shape: %d: %s %o', mfi, to.what.type, to);

            if (!f) {
                console.error('!morph.from: %d -> %o', mfi, t);
                return;
            }

            // get the 'phase' interpolation
            morphShape(f, t, phase);
        // }
        });

        //console.log('morphed: %o [%d]', m, phase);
        return m;
    }

    function animation(group, from, to, duration, done) {
        var stm = new Date().getTime();
        var phase = 0;
        var frames = 0;

        (function () {
            var tm = new Date().getTime();
            if (tm - stm < duration) {
                phase = 100 * (tm - stm) / duration;
                morph(from, to, phase);
                frames++;
                setTimeout(arguments.callee, 80);
            } else {
//console.log(dumpChart(to));
//console.log('animation: %d frames per %dms', frames, duration);
console.log('animation: '+frames+' frames per '+duration);
                if (1) {
                    group.clear();
                    draw(to, group);
                    done();
                } else {
                    if (phase >= 100) {
                        done();
                    }

                    phase = 100;
                    morph(from, to, phase);
                    setTimeout(arguments.callee, 0);
                }
            }
        })();
    }

    function animate(from, to, group, duration, done) {
        // just draw the first one
        if (from === null) {
            draw(to, group);
            done();
            return;
        }

//console.info('animate: %o', from);
        animation(group, from, to, duration, done);
    }

    this.chart = chart;
    this.axes = axes;
    this.draw = draw;
    this.animate = animate;
};

GDC = GDC || {};
GDC.charting = GDC.charting || {};

GDC.charting.chart = charting.chart;
GDC.charting.axes = charting.axes;
GDC.charting.draw = charting.draw;
GDC.charting.animate = charting.animate;

})();

