var GDC = GDC || {};
GDC.charting.chom = GDC.charting.chom || {};


GDC.charting.data.DiscreteMappingSeries = function(series, values) {
    this.getIterator = function() {
        var ctx = {
            values: values,
            index: 0,
            map: {}
        };

        return new GDC.charting.data.MapIterator(
            series.getIterator(), 
            function(v, ctx) {
                var r = ctx.map[v];
                // create v -> values mapping
                if (r === undefined) {
                    r = ctx.map[v] = ctx.values[ctx.index++];
                }
                // return the value
                return r;
            },
            ctx);
    };
};


/**
 * @class Artifact
 * @namespace GDC.charting.chom
 */
GDC.charting.chom.Artifact = function(def) {
    GDC.charting.chom.Artifact.superclass.constructor.call(this);

    this.definition = def;

    // chom artifact id - generated by the construction code
    this.id = 0;

    // styling info
    this.style = undefined;

    // label
    this.text = undefined;
};

YAHOO.lang.extend(GDC.charting.chom.Artifact, GDC.charting.layout.Visual, {
    setId: function(id) {
        this.id = id;
    },
    setStyle: function(s) {
        this.style = s;
    },
    setText: function(t) {
        this.text = t;
    },

    serialize: function(indent) {
        var s = this.id + " " + this.text;
        return s;
    }
});


/**
 * @class GridRow
 * @namespace GDC.charting.chom
 */
GDC.charting.chom.GridRow = function(def) {
    GDC.charting.chom.GridRow.superclass.constructor.call(this, def);

    this.layout = new GDC.charting.layout.HBox();

    // reflowing changes this.shape directly
    this.shape = this.layout.shape;
};

YAHOO.lang.extend(GDC.charting.chom.GridRow, GDC.charting.chom.Artifact, {
    add: function(c) {
        this.layout.add(c);
        this.children.push(c);
    },
    reflow: function(w, h) {
        return this.layout.reflow(w, h);
    },

    serialize: function(indent) {
        var s = "{row ";
        this.forEachCell(function(c) { s += c.serialize(indent); });
        s += "}";
        return s;
    },

    forEachCell: function(f) {
        this.children.forEach(f);
    }
});


/**
 * @class Grid
 * @namespace GDC.charting.chom
 */
GDC.charting.chom.Grid = function(def) {
    GDC.charting.chom.Grid.superclass.constructor.call(this, def);

    this.layout = new GDC.charting.layout.VBox();

    // reflowing changes this.shape directly
    this.shape = this.layout.shape;
};

YAHOO.lang.extend(GDC.charting.chom.Grid, GDC.charting.chom.Artifact, {
    add: function(c) {
        this.layout.add(c);
        this.children.push(c);
    },
    reflow: function(w, h) {
        return this.layout.reflow(w, h);
    },

    serialize: function(indent) {
        var s = "{grid ";
        this.forEachRow(function(r) { s += r.serialize(indent); });
        s += "}";
        return s;
    },

    forEachRow: function(f) {
        this.children.forEach(f);
    }
});


/**
 * @class Scattered
 * @namespace GDC.charting.chom
 */
GDC.charting.chom.Scattered = function(def) {
    GDC.charting.chom.Scattered.superclass.constructor.call(this, def);

    // line chart or just a scattered
    this.line = null;
};

YAHOO.lang.extend(GDC.charting.chom.Scattered, GDC.charting.chom.Artifact, {
    serialize: function(indent) {
        var s = this.line ? "{line " : "{scattered ";
        this.forEachMarker(function(m) { s += m.serialize(indent); });
        s += "}";
        return s;
    },

    forEachMarker: function(f) {
        this.children.forEach(f);
    },

    setLineStyle: function(s) {
        this.line = s;
    },

    reflow: function(w, h) {
        var d = this.definition;
        var s = this.shape;
        s.w = w;
        s.h = h;

        var size = d.series['size'];
        var propsS, ds;
        if (size === undefined) {
            propsS = {max:2, min:1, count:this.children.length};
            ds = this.line.width*2;
        } else {
            propsS = size.scaler.range;
            ds = Math.min(h,w)/((propsS.max-propsS.min)*propsS.count);
        }
        
        var xs = d.series['x'].scaler;
        var ys = d.series['y'].scaler;

        // make sure the markers fit into the designated area
        var smax = (propsS.max-propsS.min) * ds;
        xs.setMargin({beg: smax, end: smax});
        ys.setMargin({beg: smax, end: smax});

        this.children.forEach(function(e) {
            e.shape.h = e.shape.w = e.w * ds;
            e.shape.x = xs.map(e.x, w) - e.shape.w/2;
            e.shape.y = h - ys.map(e.y, h) - e.shape.h/2;

            // nest deeper
            e.reflow(e.shape.w, e.shape.h);
        });

        // we certainly have NOT changed our size in Scattered.reflow
        return null;
    },

    render: function(g, gfx) {
        // check whether there are children to render
        if (!this.children.length) {
            return null;
        }

        if (gfx) {
            g = gfx;
        } else {
            g = g.createGroup();
        }

        if (this.line) {
            var path = g.createPath();

            this.children.forEach(function(e) {
                var s = e.shape;
                var x = s.x + s.w/2;
                var y = s.y + s.h/2;
                if (path.getShape().path.length > 0)
                    path.lineTo(x,y);
                else
                    path.moveTo(x,y);

                e.render(g, e.gfx);
            });

            path.setStroke(this.line);
        } else {
            this.children.forEach(function(e) {
                e.render(g, e.gfx);
            });
        }


        var s = this.shape;
        if (s.x || s.y) {
            var m = dojox.gfx.matrix;
            g.applyTransform(m.translate(s.x,s.y));
        }

        this.gfx = g;
        return g;
    }
});


/**
 * @class Rows
 * @namespace GDC.charting.chom
 */
GDC.charting.chom.Rows = function(def) {
    GDC.charting.chom.Rows.superclass.constructor.call(this, def);

    this.layout = new GDC.charting.layout.VBox();
    // reflowing changes this.shape directly
    this.shape = this.layout.shape;
};

YAHOO.lang.extend(GDC.charting.chom.Rows, GDC.charting.chom.Artifact, {
    add: function(c) {
        this.layout.add(c);
        this.children.push(c);
    },
    reflow: function(w, h) {
        return this.layout.reflow(w, h);
    },
    reflow2: function(w, h) {
        var d = this.definition;
        var s = this.shape;
        s.w = w;
        s.h = h;

        var ss = d.series['size'].scaler;
        var bw = w / this.children.length;

        this.children.forEach(function(e) {
            e.shape.h = ss.map(e.h, h);
            e.shape.y = h - e.shape.h;
            e.shape.w = w;
            e.shape.x = 0;

            // nest deeper
            e.reflow(e.shape.w, e.shape.h);
        });

        // we certainly have NOT changed our size in Scattered.reflow
        return null;
    }
});


/**
 * @class BarChart
 * @namespace GDC.charting.chom
 */
GDC.charting.chom.BarChart = function(def) {
    GDC.charting.chom.BarChart.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.BarChart, GDC.charting.chom.Artifact, {
    reflow: function(w, h) {
        var d = this.definition;
        var s = this.shape;
        s.w = w;
        s.h = h;

        var ss = d.series['size'].scaler;
        var bw = w / this.children.length;

        this.children.forEach(function(e) {
            e.shape.h = ss.map(e.h, h);
            e.shape.w = bw;
            e.shape.x = e.shape.w * e.id;
            e.shape.y = h - e.shape.h;

            // nest deeper
            e.reflow(e.shape.w, e.shape.h);
        });

        // we certainly have NOT changed our size in Scattered.reflow
        return null;
    }
});


/**
 * @class Bar
 * @namespace GDC.charting.chom
 */
GDC.charting.chom.Bar = function(def) {
    GDC.charting.chom.Bar.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.Bar, GDC.charting.chom.Artifact, {
    render: function(g, gfx) {
        var s = this.shape;
        this.gfx = g.createRect({
                        x: s.x,
                        y: s.y,
                        width: s.w,
                        height: s.h
                   })
                    .setFill(this.color); // FIXME
        return this.gfx;
    }
});


GDC.charting.chom.Ellipse = function(def) {
    GDC.charting.chom.Ellipse.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.Ellipse, GDC.charting.chom.Artifact, {
    render: function(g, gfx) {
        var s = this.shape;
        this.gfx = g.createEllipse({cx: s.x + s.w/2, cy: s.y + s.h/2,
                                    rx: s.w/2, ry: s.h/2})
                    .setFill(this.color);
        return this.gfx;
    }
});



/**
 * @class PieSlice
 * @namespace GDC.charting.chom
 */
GDC.charting.chom.PieSlice = function(def) {
    GDC.charting.chom.PieSlice.superclass.constructor.call(this, def);
    // slice angle (degs)
    this.angle = 0;
    // slice rotation angle within the pie (degs)
    this.rotation = 0;
};

YAHOO.lang.extend(GDC.charting.chom.PieSlice, GDC.charting.chom.Artifact, {
    serialize: function(indent) {
        var s = "{slice " + this.angle + ":" + this.rotation;
        s += GDC.charting.chom.PieSlice.superclass.serialize.call(this);
        s += "}";
        return s;
    },

    render: function(g, gfx) {
        var s = this.shape;

        var r = Math.min(s.w/2,s.h/2);
        var a = this.angle;

        var x2 = Math.cos((a / 180) * Math.PI) * r;
        var y2 = Math.sin((a / 180) * Math.PI) * r;

        var p = g.createPath();
        p.setAbsoluteMode(true)
            .moveTo(0,0)
            .lineTo(r,0)
            .arcTo(r,r,0,a>180,true,{x:x2, y:y2})
            .closePath();
        p.setTransform(dojox.gfx.matrix.translate(s.w/2, s.h/2));
        p.applyTransform(dojox.gfx.matrix.rotateg(this.rotation));
        p.setFill(this.color); // FIXME

        this.gfx = p;
        return p;
    }
});


/**
 * @class Pie
 * @namespace GDC.charting.chom
 */
GDC.charting.chom.Pie = function(def) {
    GDC.charting.chom.Pie.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.Pie, GDC.charting.chom.Artifact, {
    serialize: function(indent) {
        var s = "{pie ";
        this.forEachSlice(function(slice) { s += slice.serialize(indent); });
        s += "}";
        return s;
    },

    forEachSlice: function(f) {
        this.children.forEach(f);
    },

    reflow: function(w, h) {
        var s = this.shape;
        s.w = w;
        s.h = h;

        // make the slices fill the same space
        this.children.forEach(function(e) {
            e.reflow(w, h);
        });
    }
});

