// FIXME: redeclaration...
var GDC = GDC || {};

(function(){

function Iterator(/*Array*/ resource) {
    this.res = resource;
    this.index = 0;
    this.curr = 0;

    this.hasNext = function() {
        return this.index < this.res.length;
    };
    this.next = function() {
        this.curr = this.res[this.index++];
        return this.curr;
    };
    this.peek = function() {
        return this.curr;
    };
}

function ConstIterator(/***/ value, /*int*/ len) {
    this.curr = value;
    this.index = 0;

    this.hasNext = function() {
        return len == undefined ? true : this.index < len;
    };
    this.next = function() {
        this.index++;
        return this.curr;
    };
    this.peek = function() {
        return this.curr;
    };
}

function IntsIterator(/*int*/ from,/*int*/ to) {
    this.curr = from - 1;
    this.end = to;

    this.hasNext = function() {
        return this.curr < this.end;
    };
    this.next = function() {
        return ++(this.curr);
    };
    this.peek = function() {
        return this.curr;
    };
}

function AddIterator(/*Iterator*/ i1,/*Iterator*/ i2) {
    this.i1 = i1;
    this.i2 = i2;

    this.hasNext = function() {
        return i1.hasNext() || i2.hasNext();
    };
    this.next = function() {
        var v1 = i1.hasNext() ? i1.next() : 0;
        var v2 = i2.hasNext() ? i2.next() : 0;
        return this.curr = v1 + v2;
    };
    this.peek = function() {
        return this.curr;
    };
}

function IdentityIterator(/*Iterator*/ iterator) {
    this.inner = iterator;

    this.hasNext = function() {
        return this.inner.hasNext();
    };
    this.next = function() {
        return this.inner.peek(); // YES - peek()
    };
    this.peek = function() {
        return this.inner.peek();
    };
}

function MapIterator(/*Iterator*/ iterator, /*Function*/ functor, /***/ data) {
    this.inner = iterator;
    this.curr = 0;

    this.hasNext = function() {
        return this.inner.hasNext();
    };
    this.next = function() {
        this.curr = functor(this.inner.next(), data);
        return this.curr;
    };
    this.peek = function() {
        return this.curr;
    };
}

function Series(/*Array*/ resource) {
    this.res = resource;

    this.getIterator = function() {
        return new Iterator(this.res);
    };
    this.mapToArray = function(callback, scope) {
        return this.res.map(function(node) {
            if (node.map) {
                // GDC.util.log("mapping further");
                return node.mapToArray(callback, scope);
            } else {
                // GDC.util.log("caliing callback");
                var val = callback.call(scope, node);
                // GDC.util.log(val);
                return val;
            }
        }, scope);
    };
    
    this.map = function(callback, scope) {
        return this.res.map(function(node) {
            if (node.map) {
                // GDC.util.log("mapping further");
                return new Series(node.map(callback, scope));
            } else {
                // GDC.util.log("caliing callback");
                var val = callback.call(scope, node);
                // GDC.util.log(val);
                return val;
            }
        }, scope);
    };
}

function ConstSeries(/***/ resource, /*int*/ len) {
    this.res = resource;

    this.getIterator = function() {
        return new ConstIterator(this.res, len);
    };
}

function IntsSeries(/*int*/ from, /*int*/ to) {
    this.from = from;
    this.to = to;

    this.getIterator = function() {
        return new IntsIterator(this.from, this.to);
    };
}

function AddSeries(/*Series*/ s1, /*Series*/ s2) {
    this.s1 = s1;
    this.s2 = s2;

    this.getIterator = function() {
        return new AddIterator(this.s1.getIterator(), this.s2.getIterator());
    };
}

function TransposedIterator(iteratorArray) {
    this.its = iteratorArray;

    this.hasNext = function() {
        var hn = false;
        this.its.forEach(function(it) {
            if (it.hasNext()) hn = true;
        });
        return hn;
    };
    this.next = function() {
        var d = new Array();
        this.its.forEach(function(it) {
            if (it.hasNext())
                d.push(it.next());    
            else
                d.push(null);
        });
        alert(dump(d));
        return this.curr = new Series(d);
    };
    this.peek = function() {
        return this.curr;
    };
}

function TransposedSeries(seriesArray) {
    // get iterators
    var its = new Array();
    seriesArray.forEach(function(s) {
        its.push(s.getIterator());
    });

    // construct tranposed series array 't'
    this.res = new Array();
    do {
        var d = new Array();
        hasData = false;
        its.forEach(function(it) {
            if (it.hasNext()) {
                hasData = true;
                d.push(it.next());    
            } else
                d.push(null);
        });
        if (!hasData) break;
        this.res.push(new Series(d));
    } while(true);

    this.getIterator = function() {
        return new Iterator(this.res);
    };
}



/**
 * Dependency series...
 **/

function IteratingSeries(series) {
    this.base = series;
    this.idx = -1;

    this.getIterator = function() {
        this.idx = -1;

        return new MapIterator(this.base.getIterator(), 
                               function(v, series) { series.idx++; return v; },
                               this);
    };
    this.getIndex = function() {
        return this.idx;
    };
}

function ControlledSeries(series, baseSeries) {
    this.ser = series;
    this.base = baseSeries;

    this.getIterator = function() {
        // FIXME: HACK accessing the series resource directly (needs redesign)
        if (this.ser.res[this.base.getIndex()] == undefined) {
            var x =("ConstrolledSeries: undef iterator: @" + this.base.getIndex() + "\nser.res: " + this.ser.res + "\nbase.res: " + this.base.res);
            GDC.util.error(x);
            return null;
        }
        return this.ser.res[this.base.getIndex()].getIterator();
    };
}

function InspectedSeries(series, over) {

    function inspectSeriesBy(series, over) {
        if (!over || !over.length) {
            return new SeriesProperties().inspect(series.getIterator());
        }

        var props = new SeriesProperties();
        // use a copy of the 'over' array
        var oc = new Array();
        for (var property in over) oc[property] = over[property];

        var o = oc.shift();
        var iter = o.getIterator();
        while(iter.hasNext()) {
            iter.next();

            props.merge( inspectSeriesBy(series, oc) );
        }
        return props;
    }

    function inspectSeries(series, over) {
        if (over && !YAHOO.lang.isArray(over)) {
            if (over.hasOwnProperty('getIterator')) {
                over = [over];
            } else {
                GDC.util.error("InspectedSeries: Cannot inspect over an object that is not Series instance.");
            }
        }
        return inspectSeriesBy(series, over);
    }

    this.getIterator = function() {
        return series.getIterator();
    };

    this.props = inspectSeries(this, over);
}

/**
 * Flattened series iterator contains list data items collected in
 * a DFS series hierarchy tree walk.
 **/

function FlattenIterator(series) {
    this.res = new Array();
    this.index = 0;
    this.curr = 0;

    function pushItem(container, item) {
        if (item === null || item.getIterator == undefined) {
            container.push(item);
            return;
        }
        // recursive iterator flatten
        var it = item.getIterator();
        while(it.hasNext()) {
            pushItem(container, it.next());
        }
    }

    pushItem(this.res, series);

    this.hasNext = function() {
        return this.index < this.res.length;
    };
    this.next = function() {
        this.curr = this.res[this.index++];
        return this.curr;
    };
    this.peek = function() {
        return this.curr;
    };
}

function FlattenedSeries(series) {
    this.ser = series;

    this.getIterator = function() {
        return new FlattenIterator(this.ser);
    };
}

/**
 * Spatial projection specs.
 **/
function XYZSpecs() {
    this.x = { count: 0 };
    this.y = { count: 0 };

    /**
     * Merge two axis structures.
     * @param o Other XYZSpecs internal axis structure
     **/
    function mergeAxis(t, o) {
        if (!t.count) {
            t.count = o.count;
            t.min = o.min;
            return t;
        }

        t.count += o.count;
        t.min = t.min < o.min ? t.min : o.min;
        return t;
    }

    function merge(other) {
        mergeAxis(this.x, other.x);
        mergeAxis(this.y, other.y);
    }

    // public members
    this.merge = merge;
}


/**
 * Metric inspection.
 **/

/**
 * Series inspection data representation object.
 **/
function SeriesProperties() {
    this.count = 0;

    /**
     * Conveniently merge two series properties.
     * @param other SeriesProperties instance to merge info from
     **/
    function merge(other) {
        if (!this.count) {
            this.count = other.count;
            this.sum = other.sum;
            this.min = other.min;
            this.max = other.max;
            return this;
        }

        this.count = this.count > other.count ? this.count : other.count;
        this.sum = this.sum > other.sum ? this.sum : other.sum;
        this.min = this.min < other.min ? this.min : other.min;
        this.max = this.max > other.max ? this.max : other.max;
        return this;
    }

    function mergeElement(e) {
        if (!this.count) {
            this.count = 1;
            this.sum = e;
            this.min = e;
            this.max = e;
            return e;
        }

        this.count++;
        if (e === null) return e; // neglect null values
        this.sum += e;
        this.min = this.min < e ? this.min : e;
        this.max = this.max > e ? this.max : e;
        return e;
    }

    function inspect(it) {
        if (!it || !it.hasNext()) {
            return this;
        }

        var count = 1;
        var min = it.next();
        // find the first non-null value
        while (min === null && it.hasNext()) {
            count++;
            min = it.next();
        }
        // all iterator values are 'null' -> min = 0
        if (min === null)
            min = 0;
        var max = min;
        var sum = min;
        while(it.hasNext()) {
            var e = it.next();
            count++;
            if (e === null) continue; // neglect null values
            sum += e;
            min = min<e ? min:e;
            max = max>e ? max:e;
        }

        this.count = count;
        this.sum = sum;
        this.min = min;
        this.max = max;
        return this;
    }

    // public members
    this.merge = merge;
    this.mergeElement = mergeElement;
    this.inspect = inspect;
}


function inspect(it) {
    if (! it.hasNext()) {
        return new SeriesProperties();
    }
        
    var min = it.next();
    var max = min;
    var sum = min;
    var count = 1;
    while(it.hasNext()) {
        var e = it.next();
        sum += e;
        min = min<e ? min:e;
        max = max>e ? max:e;
        count++;
    }

    return { min: min, max: max, sum: sum, count: count };
}


// FIXME: so far unused...

function NeighborSumIterator(iterator) {
    return new MapIterator(iterator,
            function(v, data) {
        var r = v + data.prev;
        data.prev = v;
        return r;
    }, {
        prev :iterator.next()
    });
}


GDC.charting = GDC.charting || {};

GDC.charting.data = {};
GDC.charting.data.Series = Series;
GDC.charting.data.Iterator = Iterator;
GDC.charting.data.MapIterator = MapIterator;

GDC.charting.data.IteratingSeries = IteratingSeries;
GDC.charting.data.ControlledSeries = ControlledSeries;

GDC.charting.data.AddSeries = AddSeries;
GDC.charting.data.TransposedSeries = TransposedSeries;

GDC.charting.data.SeriesProperties = SeriesProperties;
GDC.charting.data.InspectedSeries = InspectedSeries;

GDC.charting.data.ConstSeries = ConstSeries;
GDC.charting.data.IntsSeries = IntsSeries;

GDC.charting.data.XYZSpecs = XYZSpecs;

})();
