var GDC = GDC || {};
GDC.charting.def = GDC.charting.def || {};
GDC.charting.chom.def = GDC.charting.chom.def || {};

(function(){

 var XYZSpecs = GDC.charting.data.XYZSpecs;

GDC.charting.def.Definition = function(type) {
    /**
     * Type of the chart definition determining what chom.Artifact to use
     * (see FIXME: TODO for values).
     * @property type
     */
     this.type = type;

    /**
     * Visual styling information.
     * @property style
     */
    this.style = undefined;

    /**
     * Generated Artifact labels styling and placement configuration.
     * @property labels
     */
    this.labels = undefined;
};

YAHOO.lang.augmentObject(GDC.charting.def.Definition.prototype, {
    setStyle: function(s) {
        this.style = s;
    },
    setLabels: function(l) {
        this.labels = l;
    },

    getXYZStats: function() {
        var s = new XYZSpecs();
        s.x.count = 1;
        s.x.min = 5;
        s.y.count = 1;
        s.y.min = 5;

        s.margin = { top: 0, right: 0, bottom: 0, left: 0 };
        return s;
    }
});


GDC.charting.def.ChartDefinition = function(type) {
    GDC.charting.def.ChartDefinition.superclass.constructor.call(this, type);

    /**
     * @property what
     */
    this.what = undefined;

    /**
     * An axes series to render and to map data.Series values to the
     * target coordinate space.
     * @property axes
     */
    this.axes = [];

    /**
     * def.Series definitions.
     * @property series
     */
    this.series = {};
};

YAHOO.lang.extend(GDC.charting.def.ChartDefinition, GDC.charting.def.Definition, {
    addAxis: function(a) {
        this.axes.push(a);
    },
    setSeries: function(seriesName, series) {
        this.series[seriesName] = series;
    }
});


GDC.charting.def.Series = function(type, data, scaler) {
    this.type = type;

    /**
     * An data.Series instance with the series data.
     * @property data
     */
    this.data = data;

    /**
     * @property scaler
     */
    this.scaler = scaler;
};

YAHOO.lang.augmentObject(GDC.charting.def.Series.prototype, {
    map: function(val, availablePixels) {
        return this.scaler ? this.scaler.map(val, availablePixels) : 0;
    },

    getIterator: function(span) {
        return new GDC.charting.data.MapIterator(
            this.data.getIterator(),
            function(v, scaler) { return scaler.map(v, span); },
            this.scaler);
    }
});

GDC.charting.def.DiscreteSeries = function(data, scaler) {
    GDC.charting.def.DiscreteSeries.superclass.constructor.call(this,
        'discrete', data, scaler);
};

YAHOO.lang.extend(GDC.charting.def.DiscreteSeries, GDC.charting.def.Series);


GDC.charting.def.ContinuousSeries = function(data, scaler) {
    GDC.charting.def.DiscreteSeries.superclass.constructor.call(this,
        'continuous', data, scaler);
};

YAHOO.lang.extend(GDC.charting.def.ContinuousSeries, GDC.charting.def.Series);


GDC.charting.def.Axis = function(type) {
    // Axis type - basically its orientation
    this.type = type;

    // Unique identification used for multiple axes support
    this.id = undefined;

    // Axis' label series
    this.ticks = undefined;

    // Axis range (metric axis)
    this.range = undefined;

    // Include zero
    this.zero = false;

    // Style information
    this.style = undefined;

    // Axis title
    this.title = undefined;
};

YAHOO.lang.augmentObject(GDC.charting.def.Axis.prototype, {
    setId: function(id) {
        this.id = id;
    },

    setTickLabels: function(labels) {
        this.ticks = labels;
    },
    setIncludeZero: function() {
        this.zero = true;
    },
    setRange: function(r) {
        this.range = r;
        if (this.zero) {
            r = {
                min: this.range.min,
                max: this.range.max,
                count: this.range.count
            };
            if (r.min > 0) r.min = 0;
            if (r.max < 0) r.max = 0;
            this.range = r;
        }
        // set our scaler instance
        this.scaler = new GDC.charting.axes.Scaler(r);
    },

    setStyle: function(s) {
        this.style = s;
    },
    setTitle: function(t) {
        this.title = t;
    },

    getScaler: function() {
        return this.scaler;
    }
});

GDC.charting.def.XAxis = function() {
    GDC.charting.def.XAxis.superclass.constructor.call(this, 'x');
};
YAHOO.lang.extend(GDC.charting.def.XAxis, GDC.charting.def.Axis);

GDC.charting.def.YAxis = function() {
    GDC.charting.def.YAxis.superclass.constructor.call(this, 'y');
};
YAHOO.lang.extend(GDC.charting.def.YAxis, GDC.charting.def.Axis);


GDC.charting.def.Bar = function() {
    GDC.charting.def.Bar.superclass.constructor.call(this, 'BAR');
};
YAHOO.lang.extend(GDC.charting.def.Bar, GDC.charting.def.Definition);

GDC.charting.def.Ellipse = function() {
    GDC.charting.def.Ellipse.superclass.constructor.call(this, 'ELLIPSE');
};
YAHOO.lang.extend(GDC.charting.def.Ellipse, GDC.charting.def.Definition);

GDC.charting.def.PieSlice = function() {
    GDC.charting.def.PieSlice.superclass.constructor.call(this, 'PIESLICE');
};
YAHOO.lang.extend(GDC.charting.def.PieSlice, GDC.charting.def.Definition);


GDC.charting.def.Marker = function() {
    GDC.charting.def.Marker.superclass.constructor.call(this, 'marker');
};
YAHOO.lang.extend(GDC.charting.def.Marker, GDC.charting.def.Definition);

GDC.charting.def.Pie = function() {
    GDC.charting.def.Pie.superclass.constructor.call(this, 'pie');

    // default value
    this.what = new GDC.charting.def.PieSlice();
};

YAHOO.lang.extend(GDC.charting.def.Pie, GDC.charting.def.ChartDefinition, {
    setSizeSeries: function(series) {
        this.setSeries("size", series);
    },
    setColorSeries: function(series) {
        this.setSeries("color", series);
    }
});


GDC.charting.def.Grid = function() {
    GDC.charting.def.Grid.superclass.constructor.call(this, 'grid');
};

YAHOO.lang.extend(GDC.charting.def.Grid, GDC.charting.def.ChartDefinition, {
    setColumnSeries: function(series) {
        this.setSeries("cols", series);
    },
    setRowSeries: function(series) {
        this.setSeries("rows", series);
    },
    setCells: function(cdef) {
        this.what = cdef;
    },

    getXYZStats: function() {
        var sr = new XYZSpecs();

        var rows = this.series['rows'];
        var cols = this.series['cols'];

        var row_count = 0;
        var rows_it = rows.getIterator();
        while(rows_it.hasNext()) {
            rows_it.next();

            var col_count = 0;
            var sc = new XYZSpecs();

            var cols_it = cols.getIterator();
            while(cols_it.hasNext()) {
                cols_it.next();

                var se = this.what.getXYZStats();
                se.y.count = Math.max(sc.y.count, se.y.count);
                sc.y.count = 0;
                sc.merge( se );
                col_count++;
            }

            sc.x.count = Math.max(sr.x.count, sc.x.count);
            sr.x.count = 0;
            sr.merge(sc);
            row_count++;
        }

        sr.margin = { top: 0, right: 0, bottom: 0, left: 0 };
        return sr;
    }
});



GDC.charting.def.ScatteredChart = function() {
    GDC.charting.def.ScatteredChart.superclass.constructor.call(this, 'scattered');
};

YAHOO.lang.extend(GDC.charting.def.ScatteredChart, GDC.charting.def.ChartDefinition, {
    setXSeries: function(series) {
        this.setSeries("x", series);
    },
    setYSeries: function(series) {
        this.setSeries("y", series);
    },
    setSizeSeries: function(series) {
        this.setSeries("size", series);
    },
    setColorSeries: function(series) {
        this.setSeries("color", series);
    },
    setMarkerSeries: function(series) {
        this.setSeries("marker", series);
    },
    setMarkers: function(mdef) {
        this.what = mdef;
    },

    setLineStyle: function(s) {
        this.lineStyle = s;
    }
});


GDC.charting.def.BarChart = function() {
    GDC.charting.def.BarChart.superclass.constructor.call(this, 'bars');
};

YAHOO.lang.extend(GDC.charting.def.BarChart, GDC.charting.def.ChartDefinition, {
    setSizeSeries: function(series) {
        this.setSeries("size", series);
    },
    setColorSeries: function(series) {
        this.setSeries("color", series);
    },
    setMarkers: function(mdef) {
        this.what = mdef;
    },

    getXYZStats: function() {
        var xyz = new XYZSpecs();

        var heights = this.series['size'];
        var heights_it = heights.getIterator();
        while(heights_it.hasNext()) {
            heights_it.next();

            var exyz = this.what.getXYZStats();
            xyz.merge( exyz );

            // FIXME: get left/right margin
            // (left form the first child, right from the last one)
            xyz.margin = xyz.margin || { top: 0, right: 0, bottom: 0, left: 0 };
            xyz.margin.top = Math.max(xyz.margin.top, exyz.margin.top);
            xyz.margin.bottom = Math.max(xyz.margin.bottom, exyz.margin.bottom);
        }

        // columns do not contain any leaves in the Y axis
        xyz.y.count = 1;
        return xyz;
    }
});


GDC.charting.def.Rows = function() {
    GDC.charting.def.Rows.superclass.constructor.call(this, 'rows');
};

YAHOO.lang.extend(GDC.charting.def.Rows, GDC.charting.def.ChartDefinition, {
    setSizeSeries: function(series) {
        this.setSeries("size", series);
    },
    setColorSeries: function(series) {
        this.setSeries("color", series);
    },
    setMarkers: function(mdef) {
        this.what = mdef;
    }
});


GDC.charting.def.StackedBarChart = function() {
    GDC.charting.def.StackedBarChart.superclass.constructor.call(this, 'stackedbars');

    this.what = new GDC.charting.def.Rows();
    this.what.setMarkers(new GDC.charting.def.Bar());
};

YAHOO.lang.extend(GDC.charting.def.StackedBarChart, GDC.charting.def.ChartDefinition, {
    setSizeSeries: function(series) {
        this.setSeries("size", series);
    },
    setStackedSizeSeries: function(series) {
        this.what.setSizeSeries(series);
    },
    setStackedColorSeries: function(series) {
        this.what.colors = this.colors; // HACK FIXME
        this.what.setColorSeries(series);
    }
});



GDC.charting.axes = GDC.charting.axes || {};

GDC.charting.axes.Scaler = function(range) {
    this.range = range;

    // this.transform set to identity
    this.setTransform( function(x) { return x; } );

    this.margin = {beg: 0, end: 0};
};


YAHOO.lang.augmentObject(GDC.charting.axes.Scaler.prototype, {
   setMargin: function(m) {
       this.margin = m; // FIXME: clone?
   },
   setTransform: function(t) {
       this.transform = t;
       this.transformedRange = this.transform(this.range.max - this.range.min);
   },

   toPixel: function(x, availablePixels) {
      var m = this.margin;
      return m.beg +
             (x * (availablePixels - m.beg - m.end)) / this.transformedRange;
   },

   map: function(x, availablePixels) {
      return this.toPixel(this.transform(x-this.range.min), availablePixels);
   },

   // axis low and high pixel values
   high: function(availablePixels) {
      return this.map(this.range.max, availablePixels);
   },
   low: function(availablePixels) {
      return this.map(this.range.min, availablePixels);
   }
});


function mapSeries(series, values) {
    if (series.type === 'discrete') {
        return new GDC.charting.data.DiscreteMappingSeries(series.data, values);
    } else {
        // FIXME: map range to range (color scales etc.)
        return new GDC.charting.data.DiscreteMappingSeries(series.data, values);
    }
}


GDC.charting.chom.def.Builder = function(def) {
    this.definition = def;
};

function SetProperties(s, ctx) {
    s.setId( ctx.id );
    s.setStyle( ctx.style );
    s.setText( ctx.text );

    s.color = ctx.color;

    // used by ..reflow() later
    s.x = ctx.x;
    s.y = ctx.y;
    s.w = ctx.w;
    s.h = ctx.h;
}


GDC.charting.chom.def.MarkerBuilder = function(def) {
    GDC.charting.chom.def.MarkerBuilder.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.def.MarkerBuilder, GDC.charting.chom.def.Builder, {
    construct: function(ctx) {
        var s;
        switch (ctx.marker) {
            case "BAR":
                s = new GDC.charting.chom.Bar();
                break;
            case "ELLIPSE":
                s = new GDC.charting.chom.Ellipse();
                break;
            default:
                s = new GDC.charting.chom.Bar();
                break;
        }
        SetProperties(s, ctx);
        return s;
    }
});


GDC.charting.chom.def.BarBuilder = function(def) {
    GDC.charting.chom.def.BarBuilder.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.def.BarBuilder, GDC.charting.chom.def.Builder, {
    construct: function(ctx) {
        var s = new GDC.charting.chom.Bar();
        SetProperties(s, ctx);
        return s;
    }
});


GDC.charting.chom.def.EllipseBuilder = function(def) {
    GDC.charting.chom.def.EllipseBuilder.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.def.EllipseBuilder, GDC.charting.chom.def.Builder, {
    construct: function(ctx) {
        var s = new GDC.charting.chom.Ellipse();
        SetProperties(s, ctx);
        return s;
    }
});


GDC.charting.chom.def.PieSliceBuilder = function(def) {
    GDC.charting.chom.def.PieSliceBuilder.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.def.PieSliceBuilder, GDC.charting.chom.def.Builder, {
    construct: function(ctx) {
        var s = new GDC.charting.chom.PieSlice();
        SetProperties(s, ctx);

        s.angle = ctx.angle;
        s.rotation = ctx.rotation;
        return s;
    }
});


GDC.charting.chom.def.PieBuilder = function(def) {
    GDC.charting.chom.def.PieBuilder.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.def.PieBuilder, GDC.charting.chom.def.Builder, {
    construct: function(ctx) {
        var d = this.definition;
        var s = new GDC.charting.chom.Pie(d);
        SetProperties(s, ctx);

        var size = d.series["size"];
        var sizeit = size.data.getIterator();
        var color = d.series["color"];
        var colorsit;
        if (color !== undefined) {
            colorsit = mapSeries(color, d.colors).getIterator();
        } else {
            colorsit = new GDC.charting.data.ConstSeries(d.colors[0]).getIterator();
        }

        var builder = GDC.charting.chom.def.getBuilder(d.what);

        var da = 360 / size.data.props.sum;
        var rotation = 0;
        var n = 0;
        while(sizeit.hasNext()) {
            var ec = {
                id: n++,
                angle: da * sizeit.next(),
                rotation: rotation,
                color: colorsit.next()
            };
            // make it a square
            ec.h = ec.w;

            rotation += ec.angle;

            s.add( builder.construct(ec) );
        }
        return s;
    }
});


GDC.charting.chom.def.ScatteredBuilder = function(def) {
    GDC.charting.chom.def.ScatteredBuilder.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.def.ScatteredBuilder, GDC.charting.chom.def.Builder, {
    construct: function(ctx) {
        var d = this.definition;
        var s = new GDC.charting.chom.Scattered(d);
        SetProperties(s, ctx);

        s.setLineStyle(d.lineStyle);

        var x = d.series["x"];
        var y = d.series["y"];

        var size = d.series["size"];
        var sizeit;
        if (size !== undefined) {
            sizeit = size.data.getIterator();
        } else {
            sizeit = new GDC.charting.data.ConstSeries(1).getIterator();
        }
        var color = d.series["color"];
        var colorsit;
        if (color !== undefined) {
            colorsit = mapSeries(color, d.colors).getIterator();
        } else {
            colorsit = new GDC.charting.data.ConstSeries(d.colors[0]).getIterator();
        }
        var marker = d.series["marker"];
        var markerit;
        if (marker !== undefined) {
            markerit = mapSeries(marker, d.markers).getIterator();
        }

        var xit = x.data.getIterator();
        var yit = y.data.getIterator();
        //var textit = text.data.getIterator();

        var builder = GDC.charting.chom.def.getBuilder(d.what);
        var n = 0;
        while(xit.hasNext() && yit.hasNext()) {
            var ec = {
                id: n++,
                x: xit.next(),
                y: yit.next(),
                w: sizeit.next(),
                color: colorsit.next()
          //      text: textit.next(),
            };
            // make it a square
            ec.h = ec.w;
            // set the marker type if desired
            if (marker) ec.marker = markerit.next();

            s.add( builder.construct(ec) );
        }
        return s;
    }
});


GDC.charting.chom.def.BarChartBuilder = function(def) {
    GDC.charting.chom.def.BarChartBuilder.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.def.BarChartBuilder, GDC.charting.chom.def.Builder, {
    construct: function(ctx) {
        var d = this.definition;
        var s = new GDC.charting.chom.BarChart(d);
        SetProperties(s, ctx);

        var size = d.series["size"];
        var sizeit = size.data.getIterator();

        var color = d.series["color"];
        var colorsit;
        if (color !== undefined) {
            colorsit = mapSeries(color, d.colors).getIterator();
        } else {
            colorsit = new GDC.charting.data.ConstSeries(d.colors[0]).getIterator();
        }

        var builder = GDC.charting.chom.def.getBuilder(d.what);
        var n = 0;
        while(sizeit.hasNext()) {
            var ec = {
                id: n++,
                h: sizeit.next(),
                color: colorsit.next()
          //      text: textit.next(),
            };

            s.add( builder.construct(ec) );
        }
        return s;
    }
});


GDC.charting.chom.def.RowsBuilder = function(def) {
    GDC.charting.chom.def.RowsBuilder.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.def.RowsBuilder, GDC.charting.chom.def.Builder, {
    construct: function(ctx) {
        var d = this.definition;
        var s = new GDC.charting.chom.Rows(d);
        SetProperties(s, ctx);

        var size = d.series["size"];
        var sizeit = size.data.getIterator();

        var color = d.series["color"];
        var colorsit;
        if (color !== undefined) {
            colorsit = mapSeries(color, d.colors).getIterator();
        } else {
            colorsit = new GDC.charting.data.ConstSeries(d.colors[0]).getIterator();
        }

        var builder = GDC.charting.chom.def.getBuilder(d.what);
        var n = 0;
        while(sizeit.hasNext()) {
            var ec = {
                id: n++,
                h: sizeit.next(),
                color: colorsit.next()
          //      text: textit.next(),
            };

            var e = builder.construct(ec);
            e.flex = ec.h;
            s.add(e)
        }
        return s;
    }
});


GDC.charting.chom.def.GridBuilder = function(def) {
    GDC.charting.chom.def.GridBuilder.superclass.constructor.call(this, def);
};

YAHOO.lang.extend(GDC.charting.chom.def.GridBuilder, GDC.charting.chom.def.Builder, {
    construct: function(ctx) {
        var d = this.definition;
        var s = new GDC.charting.chom.Grid(d);
        SetProperties(s, ctx);

        var builder = GDC.charting.chom.def.getBuilder(d.what);
        if (!builder) {
            GDC.util.error("Grid cell definition not valid");
            return s;
        }

        var cols = d.series["cols"];
        var rows = d.series["rows"];

        var xyz;

        var r = 0;
        var rowsit = rows.data.getIterator();
        while(rowsit.hasNext()) {
            var rv = rowsit.next();
            var row = new GDC.charting.chom.GridRow(d);

            var c = 0;
            var colsit = cols.data.getIterator();
            while(colsit.hasNext()) {
                var ec = {
                    id: r +":"+ c,
                    row: rv,
                    col: colsit.next()
                };

                var cell = builder.construct(ec);
                xyz = d.what.getXYZStats(); // perhaps builder.getXYZStats()?

                c++;
                cell.flex = xyz.x.count;
                row.add(cell);
            }

            r++;
            row.flex = xyz.y.count;
            s.add(row);
        }
        return s;
    }
});


})();

GDC.charting.chom.def.getBuilder = function(definition) {
        var b = undefined;
        if (definition instanceof GDC.charting.def.ScatteredChart) {
            b = new GDC.charting.chom.def.ScatteredBuilder(definition);
        } else if (definition instanceof GDC.charting.def.BarChart) {
            b = new GDC.charting.chom.def.BarChartBuilder(definition);
        } else if (definition instanceof GDC.charting.def.Grid) {
            b = new GDC.charting.chom.def.GridBuilder(definition);
        } else if (definition instanceof GDC.charting.def.StackedBarChart) {
            b = new GDC.charting.chom.def.BarChartBuilder(definition);
        } else if (definition instanceof GDC.charting.def.Rows) {
            b = new GDC.charting.chom.def.RowsBuilder(definition);
        } else if (definition instanceof GDC.charting.def.Marker) {
            b = new GDC.charting.chom.def.MarkerBuilder(definition);
        } else if (definition instanceof GDC.charting.def.Bar) {
            b = new GDC.charting.chom.def.BarBuilder(definition);
        } else if (definition instanceof GDC.charting.def.Ellipse) {
            b = new GDC.charting.chom.def.EllipseBuilder(definition);
        } else if (definition instanceof GDC.charting.def.Pie) {
            b = new GDC.charting.chom.def.PieBuilder(definition);
        } else if (definition instanceof GDC.charting.def.PieSlice) {
            b = new GDC.charting.chom.def.PieSliceBuilder(definition);
        }

        return b;
};


