(function(){
    var Series = GDC.charting.data.Series;
    var ConstSeries = GDC.charting.data.ConstSeries;
    var IntsSeries = GDC.charting.data.IntsSeries;
    var IteratingSeries = GDC.charting.data.IteratingSeries;
    var ControlledSeries = GDC.charting.data.ControlledSeries;
    var InspectedSeries = GDC.charting.data.InspectedSeries;

    /** markers **/

    function getMarker( type, opts ) {
        return {
            type: type,
            style: GDC.util.clone( opts.style ),
            labels: GDC.util.clone( opts.labels )
        };
    }

    function getPieSeries(aMetric, rMetric, opts) {
	    return {
            width: aMetric.data,
            radius: rMetric.data
            //color: colorMetric,
            //text: textMetric
        };
    }

    function getPieMarker(aSeries, rSeries, opts) {
        var pie =
        {
            type: "pie",
            series: getPieSeries(aSeries, rSeries, opts),
            what: getMarker( "PIESLICE", opts.markers )
        };
        return pie;
    }

    function getRowsSeries(series, opts) {
	    return {
            size: series.data,
            text: series.data
            //colors: colorSeries
        };
    }

    function getRowsMarker(what, series, opts) {
        var rows =
        {
            type: "rows",
            series: getRowsSeries(series, opts.colorScheme),
            what: what
        };
        return rows;
    }

    /** axes **/

    function getAxis( kind, type, series, opts ) {
        var a = {
            type: type,
            title: series.name,
            //ticks: series,
            //range: series.props,
            style: GDC.util.clone(opts.style)
		};

        if (kind === "metric") {
            a.range = series.data.props;
        } else {
            a.ticks = series.data;
        }
        return a;
    }

    /** series **/

    function getColumnsSeries(metric, colorScheme) {
	    return {
            height: metric.data,
            colors: new ConstSeries(["lightskyblue"]) // FIXME!!!
            // text: metric
        };
    }

    function getScatteredSeries(xMetric, yMetric, opts) {
        // if x is attribute
        var x = new IntsSeries(0, new InspectedSeries(xMetric.data).props.count-1);
	    return {
            x: x,
            y: yMetric.data,
            //size: sizeMetric,
            //color: colorMetric,
            color: new ConstSeries(["lightskyblue"]) // FIXME!!!
            //text: textMetric
        };
    }

    /** tooltip **/
    function getToolTip(xMetric, yMetric, opts) {
	    return {
            style: GDC.util.clone(opts.style),
            content: [
                { data: [xMetric.data],
                  names: [""]
                }
            ]
        };
    }

    /** charts **/

    function getBarChart(xSeries, ySeries, opts) {
        var columns =
        {
            type: "columns",
            pad:  0,
            axes: [
                getAxis("attrib", "x", xSeries, opts.xaxis),
                getAxis("metric", "y", ySeries, opts.yaxis)
            ],
            series: getColumnsSeries(ySeries, opts.colorScheme),
            what: getMarker( "BAR", opts.markers )
        };
        columns.tooltip = getToolTip(xSeries, ySeries, opts.tooltip);
        return columns;
    }

    function getStackedBarChart(xSeries, ySeries, stackSeries, opts) {
        var bars = getBarChart(xSeries, ySeries, opts);
        bars.what = getRowsMarker( getMarker( "BAR", opts.bars ),
                                   stackSeries, opts.markers );
        return bars;
    }

    function getOverlay(xSeries, ySeries, opts) {
        var overlay =
        {
            type: "overlay",
            axes: [
                getAxis("attrib", "x", xSeries, opts.xaxis),
                getAxis("metric", "y", ySeries, opts.yaxis)
            ],
            series: []
        };
        overlay.tooltip = getToolTip(xSeries, ySeries, opts.tooltip);
        return overlay;
    }

    function getScatteredChart(xSeries, ySeries, opts) {
        var scattered =
        {
            type: "scattered",
            style: GDC.util.clone(opts.style),
            axes: [
                getAxis("attrib", "x", xSeries, opts.xaxis),
                getAxis("metric", "y", ySeries, opts.yaxis)
            ],
            series: getScatteredSeries(xSeries, ySeries, opts),
            what: getMarker( "ELLIPSE", opts.markers )
        };
        scattered.tooltip = getToolTip(xSeries, ySeries, opts.tooltip);
        return scattered;
    }

    function getLineChart(xSeries, ySeries, opts) {
        var scattered = getScatteredChart(xSeries, ySeries, opts);
        scattered.type = "line";
        return scattered;
    }

    function getPieChart(xSeries, ySeries, pies, opts) {
        var pies = getScatteredChart(xSeries, ySeries, opts);
        pies.what = getPieMarker( pies, null, opts.markers );
        return pies;
    }

    function attrCombinations( a, prefix, overSeries, functor ) {
        if (!overSeries.length) {
            functor(a, prefix);
            return;
        }

        // clone 'over'
        var over = []; overSeries.forEach(function(i) { over.push(i); });
        var o = over.shift();

        var it = o.getIterator();
        while(it.hasNext()) {
            var v = it.next();
            attrCombinations(a, prefix + " / " + v, over, functor);
        }
    }

    function rowMetricCombinations(a, i, metrics, overSeries) {
        if (!overSeries.length) {
            function runMetrics(si) {
                metrics.forEach(function(m) {
                    a[si] = a[si] || [];
                    var mit = m.data.getIterator();
                    while(mit.hasNext()) {
                        a[si].push( mit.next());
                    }
                    si++;
                });
                return si;
            }

            return runMetrics(i);
        }

        // clone 'over'
        var over = []; overSeries.forEach(function(ovi) { over.push(ovi); });
        var o = over.shift();

        var it = o.getIterator();
        while(it.hasNext()) {
            it.next(); // just iterate
            i = rowMetricCombinations(a, i, metrics, over);
        }
        return i;
    }

    function colMetricCombinations(a, i, metrics, overSeries) {
        if (!overSeries.length) {
            function runMetrics(si) {
                metrics.forEach(function(m) {
                    var mit = m.data.getIterator();
                    while(mit.hasNext()) {
                        a[si] = a[si] || [];
                        a[si].push( mit.next());
                        si++;
                    }
                });
                return si;
            }

            return runMetrics(i);
        }

        // clone 'over'
        var over = []; overSeries.forEach(function(ovi) { over.push(ovi); });
        var o = over.shift();

        var it = o.getIterator();
        while(it.hasNext()) {
            it.next(); // just iterate
            i = colMetricCombinations(a, i, metrics, over);
        }
        return i;
    }

    function metricCombinations(a, overSeries, functor) {
        if (!overSeries.length) {
            return functor(a);
        }

        // clone 'over'
        var over = []; overSeries.forEach(function(i) { over.push(i); });
        var o = over.shift();

        var it = o.getIterator();
        while(it.hasNext()) {
            it.next(); // just iterate
            metricCombinations(a, over, functor);
        }
    }

    function getChartDefinition(type, cols, rows, metrics) {
        var col = ['lightcoral', 'lightskyblue', 'lightgreen',
                   'lightcoral', 'lightskyblue', 'lightgreen',
                   'lightcoral', 'lightskyblue', 'lightgreen',
                   'lightcoral', 'lightskyblue', 'lightgreen',
                   'lightcoral', 'lightskyblue', 'lightgreen',
                   'lightcoral', 'lightskyblue', 'lightgreen',
                   'lightcoral', 'lightskyblue', 'lightgreen'
                  ];

        var axs = {size: 15, fill: 'gray', rotate: -70};
        var ays = {size: 15, fill: 'black'};

        var opts = {
            // scattered line width (size of markers/3)
            style: { width: 3, color: 'lightcoral' },
            xaxis: { style: axs },
            yaxis: { style: ays },
            markers: { style: {stroke: null, fill: 'lightskyblue', r: 1},
                       labels: {size: 15, fill: 'black'} },
            tooltip: { style: axs }
        };

        var sectionColumns = 0;
        if (metrics.length && metrics[0].section === 'columns') {
            sectionColumns = 1;
        }

        // all attributes
        var attrStructs = [];
        cols.forEach(function(c){ attrStructs.push(c); });
        rows.forEach(function(r){ attrStructs.push(r); });

        var cs = [], rs = [];
        cols.forEach(function(c){ cs.push(c.data); });
        rows.forEach(function(c){ rs.push(c.data); });

        // cut the drill attribute
        var ss = attrStructs.pop();

        // get an array of attrib Series
        var xTitle = '';
        var attrs = [];
        attrStructs.forEach(function(s){
            attrs.push(s.data);
            xTitle += " / " + s.name;
        });
        xTitle = xTitle.substring(3);

        var ss = ss.data;

        var xsa = [];
        attrCombinations(xsa, '', cs, function(a, prefix) { a.push(prefix.substring(3)); });

        var ysa = [];
        var metricTooltip = {
            names: ["", null],
            data: []
        }
        if (sectionColumns) {
            // column section metrics are drilled up-to the last row attribute
            // -> need to remove it
            var drill = rs.pop();
            attrCombinations(metricTooltip.data, '', rs, function(a, prefix) {
                metrics.forEach(function(m){
                    if (drill) {
                        var dit = drill.getIterator();
                        while(dit.hasNext()) {
                            a.push(m.name + prefix + " / " + dit.next());
                        }
                    } else {
                        a.push(m.name + prefix);
                    }
                });
            });

            metricCombinations(ysa, cs, function(a) {
                colMetricCombinations(a, 0, metrics, rs);
            });
        } else {
            attrCombinations(metricTooltip.data, '', rs, function(a, prefix) {
                metrics.forEach(function(m){
                    a.push(m.name + prefix);
                });
            });

            // row section metrics are drilled up-to the last column attribute
            // -> need to remove it
            cs.pop();
            metricCombinations(ysa, cs, function(a) {
                rowMetricCombinations(a, 0, metrics, rs);
            });
        }

        var xs = { name: xTitle,
                   data: new Series(xsa) };

        var yTitle = '';
        metrics.forEach(function(m){
            yTitle += " / " + m.name;
        });
        yTitle = yTitle.substring(3);

        var yms = [];
        var colIndex = 0;
        var overlays = [];
        ysa.forEach(function(y) {
            var yy = {
                name: yTitle,
                data: new InspectedSeries(new Series(y), attrs)
            };

            // change color for metrics
            opts.style.color = col[colIndex];
            opts.markers.style.fill = col[colIndex];
            colIndex++;

            yms.push(yy.data);

            overlays.push(getScatteredChart(xs, yy, opts));

            // set to the appropriate chart type
            switch(type) {
                case 'scattered':
                case 'line':
                    overlays[overlays.length-1].type = type;
                default:;
            }
        });

        var ySeriesIndexes = new IteratingSeries(new IntsSeries(0, ysa.length-1));
        var ySeries = new ControlledSeries(new Series(yms), ySeriesIndexes);

        var ys = { name: yTitle,
                   data: new InspectedSeries(ySeries, ySeriesIndexes) };

        var overlay = getOverlay(xs, ys, opts);
        overlay.series = new Series(overlays);

        metricTooltip.data = [ new Series(metricTooltip.data), ySeriesIndexes ];
        overlay.tooltip.content = [
            metricTooltip, 
            { names: ["", ""],
              data: [ySeries, overlay.tooltip.content[0].data[0]] }
        ];

        GDC.util.log("xs: ", xs);
        GDC.util.log("ys: ", ys);

        switch(type) {
            case 'bar':
                axs.align = "center";
                return getBarChart(xs, ys, opts);
            case 'scattered':
            case 'line':
                return overlay;

            // stacked
            case 'stackedbar':
                return getStackedBarChart(xs, ys, ss, opts);
            case 'pie':
                return getPieChart(xs, ys, ss, opts);
        }
    }

    function displayChart(elemId, type, data) { 

        function ToNumberSeries(series) {
            this.base = series;

            this.getIterator = function() {
                return new GDC.charting.data.MapIterator(this.base.getIterator(), 
                        //function(v) { return v.replace(/[^0-9,\.]/g, "") - 0; }, this);
                        function(v) { return v - 0; }, this);
            }
        }

        function getBBox(elemId) {
            var r = YAHOO.util.Dom.getRegion(document.getElementById(elemId));
            width = r.right - r.left;
            height = r.bottom - r.top;
            return {w: width, h: height};
        }
            
        // columns, rows, metrics
        var Cs = data.getColumnAttributes();
        var Rs = data.getRowAttributes();
        var Ms = data.getMetricSeries();
        Ms.forEach(function(m) { m.data = new ToNumberSeries( m.data ) });
            
        // get the chart definition tree
        var def = getChartDefinition(type, Cs, Rs, Ms);
        GDC.util.log(def);

        // ! at this point the element[id == elemId] must not
        //   be display:none
        var bb = getBBox(elemId);
        var s = new GDC.charting.gfx.Surface(elemId, bb.w, bb.h).render();
        var c = GDC.charting.chart(def, bb.w, bb.h, s);
        GDC.charting.axes(c, s);
        GDC.charting.draw(c, s);
    }


    GDC = GDC || {};
    GDC.charting = GDC.charting || {};

    GDC.charting.getChartDefinition = getChartDefinition;
    GDC.charting.displayChart = displayChart;
})();
