dojo.require("dojo.colors"); // to have all the CSS3 color definitions 
dojo.require("dojox.gfx"); 

GDC.charting.tr = new function() {
    var m = dojox.gfx.matrix;

    this.move = function(what, x, y) {
        if (what === null) return null;
        what.applyTransform(m.translate(x,y));
        return what;
    };

    this.rotate = function(what, angle) {
        if (what === null) return null;
        what.applyTransform(m.rotateg(angle));
        return what;
    };

    this.scale = function(what, dx, dy) {
        if (what === null) return null;
        what.applyTransform(m.scale(dx, dy));
        return what;
    };
};

GDC.charting.gfx = new function () {
    var tr = GDC.charting.tr;

    function setStyle(ctx) {
        ctx.gfx.setStroke(ctx.s);
        ctx.gfx.setFill(ctx.f);
    }

    function getLabelMetrics(ctx, opts, extent) {
        opts = opts || {};
        opts.halign = opts.halign || 'center';
        opts.valign = opts.valign || 'middle';

        var shape = { x: 0, y: 0 };
        var margin = { top: 0, right: 0, bottom: 0, left: 0 };

        switch(opts.halign) {
            case 'center':
                shape.x = (ctx.w-extent.w)/2-extent.w/extent.h/2;
                margin.right = (extent.w-ctx.w)/2-extent.w/extent.h/2;
                margin.right = margin.right > 0 ? margin.right : 0;
                margin.left = margin.right;
                break;
            case 'right':
                shape.x = ctx.w+extent.w/8;
                margin.right = extent.w/4 + extent.w;
                break;
        }

        switch(opts.valign) {
            case 'below':
                shape.y = ctx.h + extent.h;
                margin.bottom = extent.h + extent.h/3;
                break;
            case 'bottom':
                shape.y = ctx.h - extent.h/3;
                break;
            case 'middle':
                shape.y = (ctx.h-extent.h)/2 + extent.h - extent.h/4;
                break;
            case 'top':
                shape.y = extent.h;
                break;
            case 'above':
                shape.y = -extent.h/5;
                margin.top = extent.h;// + Math.ceil(extent.h/5);
                break;
        }

        margin.top = Math.round(margin.top);
        margin.bottom = Math.round(margin.bottom);
        margin.left = Math.round(margin.left);
        margin.right = Math.round(margin.right);

        shape.x = Math.round(shape.x);
        shape.y = Math.round(shape.y);

        //console.log('margin %o (%o)[%o]', margin, opts, extent);
        return {shape: shape, margin: margin};
    }

    function Gfx( surface ) {
        var x = surface.gdc = {};

        // create and hide the measurement group
        x.measurement = surface.createGroup();
        x.measurement.getNode().style.visibility='hidden';

        /**
         * Get a text extent (basically bounding box in other words.
         */
        function getTextExtent(text, size) {
            x.measuredText = new Text(
                    x.measurement, x.measuredText, text, size).render();
            return x.measuredText.getExtent();
        }

        /**
         * Get bounding box out of a dojox.gfx.shape. There are some shapes
         * that simply do not implement the getBoundingBox() method. Therefore
         * we go DFS inside of such shape and get the bounding box from inside.
         * This is possible and perhaps even more accurate for markers.
         */
        function getBBox(gfx) {
            var bb = gfx.getTransformedBoundingBox();
            if (!bb && gfx.children) {
                gfx.children.forEach(function(e) {
                    ebb = getBBox(e);
                    if (ebb) {
                        bb = ebb;
                    }
                });
            }
            return bb;
        }

        this.getBBox = getBBox;
        this.getTextExtent = getTextExtent;
    }

    function Surface(name, w, h) {
        this.render = function() {
            var surface = new dojox.gfx.createSurface(document.getElementById(name), w, h);
            return surface;
            //var grp = surface.createGroup();
            //return grp; //tr.move(tr.scale(grp, 1,-1), 0, -h); // fix the coordinate system bottom left
        }
    }

    function Group(ctx) {
        this.render = function() {
            return ctx.group.createGroup();
        }
    }

    function Labeled(ctx, g, renderMarker, w, h, t, opts) {
        this.render = function() {
            var text = null;
            var marker = null;

            if (! g) {
                var oldGroup = ctx.group;

                ctx.group = g = ctx.group.createGroup();
                marker = renderMarker(ctx, null);

                ctx.group = g;

                // BUG! in dojox.gfx:
                //   - reparenting texts (possibly any element) causes
                //     IE to fail (this basically means tha the transformation
                //     matrix shuffling in dojox.gfx.shape.removeShape() and
                //     dojox.gfx.shape.add() doesn't work completely

                //g.add(txt.removeShape(true));
                //g.add(txt);
                //var text = txt;

                t = t || ''; // normalize
                text = new Text(ctx.group, null, t, opts.style.size).render();

                g.setStroke = function(s) {
                    if (marker) marker.setStroke(s);
                    //text.setStroke({color:'gray', stroke:1});
                    return this; };
                g.setFill = function(f) {
                    if (marker) marker.setFill(f);
                    text.setFill(opts.style.fill);
                    return this; };

                ctx.group = oldGroup;

                if (w <= 0 || h <= 0) {
                    g.rawNode.style.visibility='hidden';
                    return g;
                }

            } else {
                if (w <= 0 || h <= 0) {
                    g.rawNode.style.visibility='hidden';
                    return g;
                }

                marker = null;
                if (g.children.length > 1) {
                    g.children[0] = renderMarker(ctx, g.children[0]);
                    text = g.children[1];
                } else {
                    text = g.children[0];
                }

                if (text.getShape().text !== t) {
                    t = t || ''; // normalize
                    text = new Text(ctx.group, text, t, opts.style.size).render();
                }
            }

            // make it visible if it wasn't before
            if ( g.rawNode.style.visibility == 'hidden')
                g.rawNode.style.visibility='visible';

            // note! the text must be visible otherwise the
            // browser cannot give out the bounding box which is used
            // in the getTextWidth() method)
            if (opts) {
                // font size in pixels
                //opts.style.size = opts.style.size || 16;  // default font size
                //text.setFont({size: opts.style.size + 'px'});

                var extent = text.getExtent();
                var metrics = getLabelMetrics(ctx, opts, extent);
                text.setShape(metrics.shape);

                // hide the text if it doesn't fit into the marker
                if (extent.h > ctx.h - extent.h/2)
                    text.rawNode.style.visibility='hidden';
            }

            return g;
        }
    }

    function Ellipse(ctx, g, rx, ry) {
        this.render = function() {
            if (! g) {
                g = ctx.group.createEllipse();
            }
            rx= Math.round( rx);
            ry= Math.round( ry);

            // necessary for VML (invalid string init Vector2D error)
            // as it does not support non-positive width and height nodes
            if (rx <= 0 || ry <= 0) {
                g.rawNode.style.visibility='hidden';
                return g;
            }

            // only meaningful sizes get here
            g.setShape({cx:rx, cy:ry, rx:rx, ry:ry});

            // make it visible if it wasn't before
            if ( g.rawNode.style.visibility == 'hidden')
                g.rawNode.style.visibility='visible';
            return g;
        }
    }
    function Line(ctx, w, h) {
        this.render = function() {
            return ctx.group.createLine({x1:0, y1:0, x2:w, y2:h});
        }
    }
    function Bar(ctx, g, w, h, r) {
        this.render = function() {
            if (! g) {
                g = ctx.group.createRect();
            }
            h= Math.round( h);
            w= Math.round( w);

            if (w <= 0 || h <= 0) {
                g.rawNode.style.visibility='hidden';
                return g;
            }

            g.setShape({x:0, y:0, width:w, height:h, r:r!==undefined ? r : w/10});

            // make it visible if it wasn't before
            if ( g.rawNode.style.visibility == 'hidden')
                g.rawNode.style.visibility='visible';
            return g;
        }
    }

    function MultiText(factory, lines, size) {
        this.render = function(g) {
            var sizePx = size + 'px';

            if (g) {
                g.setShape({text:lines});

                /*
                var sizePx = height + 'px';
                if (g.getFont().size !== sizePx) {
                    g.setFont({size: sizePx});
                }
                */

                return g;
            }

            g = factory.createGroup();
            var offset = size;
            lines.forEach(function(l) {
                var lt = new Text(g, null, l, size).render();
                lt.setFont({size: sizePx});
                    //.setStroke({color: 'gray', width: 0});
                tr.move(lt, 0, offset);

                offset += size + size/6;
            });

            g.getExtent = function(s) {
                if (!g.children)
                    return {w: 0, h: 0};

                var e = g.children[0].getExtent();
                e.h = 0;
                g.children.forEach(function(ch) {
                    var ge = ch.getExtent();
                    e.w = Math.max(e.w, ge.w);
                    e.h += ge.h + size/6;
                });
                return e;
            }
            g.setFill = function(f) {
                g.children.forEach(function(ch) {
                    ch.setFill(f);
                });
                return g;
            }
            g.setShape = function(s) {
                g.children.forEach(function(ch, i) {
                    ch.setShape({text: s.text[i]});
                });
                return g;
            }
            return g;
        }
    }

    function Text(factory, g, t, size) {
        var height = size;

        function getExtent() {
            // TODO: dojox.gfx bug? where it tries to get the text node
            // not counting on the posibility of there not being one
            if (g.getShape().text === '') {
                return {w:0, h:height};
            }

            //var text = this.children[0];
            //text.setFont({size: '28px'});
            try {
                return {w:g.getTextWidth(), h:height};
            } catch(exception) {
                console.error("getTextWidth: %o", exception);
                return {w:-1, h:height};
            }
        }
        function getExtentAdjustSize(t, widthBestFit) {
            var MIN = 14;
            if (!t.length) return {w:0, h:MIN};

            // FIXME? This is just a wild guess on how to calculate
            // the closest pixel font height based on the supplied
            // width to fit the text into and the text itself
            var th = Math.round(widthBestFit / t.length * 2.5);
            if (th<MIN) th=MIN;

            var text = this.children[0];

            text.setFont({size: th + 'px'});
            var width = text.getTextWidth();
            //console.log("1txt %d->%d", th, width);
            while (width > widthBestFit && th>MIN) {
                //console.log("2txt %d->%d", th, width);
                text.setFont({size: --th + 'px'});
                width = text.getTextWidth();
            }
            return {w:width, h:th};
        }

        this.render = function() {
            if (g) {
                g.setShape({text:t});

                var sizePx = height + 'px';
                if (!g.getFont() || g.getFont().size != sizePx) {
                    g.setFont({size: sizePx});
                }

                // need to reassign as it uses 'height' from the current
                // closure here
                g.getExtent = getExtent;
                return g;
            }

            g = factory.createText({text:t});

            /*
            // construct a text object and invert its Y axis as we
            // use different coordinate system
            g = factory.createGroup();
            //tr.scale(g.createText({text:t}), 1, -1);
            g.createText({text:t});
            // enclosed in a group and redirect the property operation
            // to the text node others are done on top of the enclosing
            // group (transformations behave differently for the invert
            // Y axis coordinate system
            g.getTextWidth = function() {
                return this.children[0].getTextWidth(); };
            g.setStroke = function(s) {
                this.children[0].setStroke(s); return this; };
            g.setFill = function(f) {
                this.children[0].setFill(f); return this; };
            g.setFont = function(f) {
                this.children[0].setFont(f); return this; };
            g.getShape = function(s) {
                return this.children[0].getShape(); };
            g.setShape = function(s) {
                // invert the Y axis value
                //if (s.y !== undefined) s.y = -s.y;
                this.children[0].setShape(s); return this; };
            */
            g.getExtent = getExtent;

            // set the font size
            if (height) {
                g.setFont({size: height + 'px'});
            }
            return g;
        }
    }
    function Path(ctx) {
        this.render = function() {
            return ctx.group.createPath();
        }
    }
    function PieSlice(ctx, g, a, r) {
        this.render = function() {
            if (! g) {
                if (a === 360) {
                    g = ctx.group.createEllipse();
                    r = Math.round( r);
                    // FIXME? what about r==0
                    g.setShape({cx:0, cy:0, rx:r, ry:r});
                    return g;
                } else {
                    g = ctx.group.createPath();
                }
            } else {
                // FIXME! TODO Animation: transitions between
                // a slice and a full bubble cannot work at this point

                g.setShape({path:''});
            }

            if (r <= 0 || a <= 0) {
                g.rawNode.style.visibility='hidden';
                return g;
            }

            var x2 = Math.cos((a / 180) * Math.PI) * r;
            var y2 = Math.sin((a / 180) * Math.PI) * r;
            r = Math.round(r);
            x2 = Math.round(x2);
            y2 = Math.round(y2);
            g.setAbsoluteMode(true)
                .moveTo(0,0)
                .lineTo(r,0)
                .arcTo(r,r,0,a>180,true,{x:x2, y:y2})
                .closePath();

            // make it visible if it wasn't before
            if ( g.rawNode.style.visibility == 'hidden')
                g.rawNode.style.visibility='visible';
            return g;
        }
    }

    this.setStyle = setStyle;
    this.getLabelMetrics = getLabelMetrics;

    this.Gfx = Gfx;

    this.Surface = Surface;
    this.Group = Group;
    this.Labeled = Labeled;

    this.Line = Line;
    this.Bar = Bar;
    this.Ellipse = Ellipse;
    this.MultiText = MultiText;
    this.Text = Text;
    this.Path = Path;

    this.PieSlice = PieSlice;
}
