var GDC = GDC || {};
GDC.charting.view = GDC.charting.view || {};

(function(){

var Box = GDC.charting.layout.Box;
var VBox = GDC.charting.layout.VBox;
var HBox = GDC.charting.layout.HBox;


GDC.charting.view.ChartWithAxesView = function(chart) {
    GDC.charting.view.ChartWithAxesView.superclass.constructor.call(this);
    
    this.leftAxes = new HBox();
    this.bottomAxes = new VBox();

    this.layout = new VBox();
    // reflowing changes this.shape directly
    this.shape = this.layout.shape;
    // add it as a child (for the ContainerBox.render() to work)
    this.add(this.layout); this.layout.flex = 1;

    var b = new HBox();
    this.layout.add(b); b.flex = 1;

    b.add( this.leftAxes ); // fixed width left axes
    b.add( chart ); chart.flex = 1;

    // fixed height bottom of the chart
    b = new HBox();
    this.layout.add(b);

    this.padding = new Box();
    b.add( this.padding ); // padding (set with to the this.leftAxes)
    b.add( this.bottomAxes ); this.bottomAxes.flex = 1;
};

YAHOO.lang.extend(GDC.charting.view.ChartWithAxesView,
                  GDC.charting.layout.Visual, {
    reflow: function(w, h, ctx) {
        return this.layout.reflow(w, h, ctx);
    },

    addLeftAxis: function(a) {
        this.leftAxes.add(a);
    },
    addBottomAxis: function(a) {
        this.bottomAxes.add(a);
    },

    calculate: function(w, h, ctx) {
        // 1. need to have the bottomAxes in and their heights set up.
        // 2. recalculate to get the leftAxes width (from their font size and
        //    height /available pixels/)
        this.reflow(w, h, ctx);
        // 3. set padding to the width of leftAxes
        this.padding.shape.w = this.leftAxes.shape.w;
        // 4. reflow with the bottomAxes height and leftAxes width set
        this.reflow(w, h, ctx);
        // 5. now the chart area should be accurate
    }
});



GDC.charting.view.GridWithAxesView = function(gridChom) {
    GDC.charting.view.GridWithAxesView.superclass.constructor.call(this, gridChom);
    function getFirst(arr, functor) {
        if (!arr || !arr.length)
            return undefined;

        var l = arr.length;
        while(l--) {
            if (functor(arr[l])) {
                return arr[l];
            }
        }
        return undefined;
    }
    
    function CollectRowAxes( chom ) {
        var d = chom.definition;
        if (!d) {
            // this happens if the chom is just a layout.Box e.g.
            return null;
        }

        // / this' choms own axis /
        var ayDef = getFirst(d.axes,
                             function(a) { return a.type === 'y'; });
        var ga = ayDef ? new GDC.charting.view.LabeledYAxis(ayDef) : null;

        if (! (chom instanceof GDC.charting.chom.Grid)) {
            return ga;
        }

        var rows = d.series['rows'];
        var rowsit = rows.data.getIterator();

        // / lower level grid axes /
        var yas = new GDC.charting.layout.VBox();
        chom.children.forEach(function(row) {
            rowsit.next();

            // set the columns iterator to the first column
            // note: this is necessary for the nested axes views
            // to display correct data (dependent on the iterators)
            d.series['cols'].data.getIterator().next();

            // get axes from the first column
            if (!row.children.length) {
                return;
            }

            var ya = CollectRowAxes( row.children[0] );
            if (ya) {
                ya.flex = row.flex;
                yas.add(ya);
            }
        });

        // HACK - layout doesn't handle opposite axis size adoption
        if (yas.children.length) {
            yas.shape.w = yas.children[0].shape.w;

            // | this' choms own axis | lower level grid axes |
            if (ga) {
                var gyas = new GDC.charting.layout.HBox();
                gyas.add(ga);
                gyas.add(yas);
                return gyas;
            }
            return yas;
        }
        return ga;
    }

    function CollectColAxes( chom ) {
        var d = chom.definition;
        if (!d) {
            // this happens if the chom is just a layout.Box e.g.
            return null;
        }

        // | this' choms own axis |
        var axDef = getFirst(d.axes,
                             function(a) { return a.type === 'x'; });
        var ga = axDef ? new GDC.charting.view.LabeledXAxis(axDef) : null;

        if (! (chom instanceof GDC.charting.chom.Grid)) {
            return ga;
        }

        // set the rows iterator to the first row
        // note: this is necessary for the nested axes views
        // to display correct data (dependent on the iterators)
        d.series['rows'].data.getIterator().next();
        var cols = d.series['cols'];
        var colsit = cols.data.getIterator();

        // | lower level grid axes |
        var xas = new GDC.charting.layout.HBox();
        if (chom.children.length) {
            chom.children[0].children.forEach(function(col) {
                colsit.next();

                var xa = CollectColAxes( col );
                if (xa) {
                    xa.flex = col.flex;
                    xas.add(xa);
                }
            });
        }

        // HACK - layout doesn't handle opposite axis size adoption
        if (xas.children.length) {
            //xas.shape.h = 20; // HMM... ?!? why here?
            xas.shape.h = xas.children[0].shape.h;

            // / lower level grid axes / this' choms own axis /
            if (ga) {
                var gxas = new GDC.charting.layout.VBox();
                gxas.add(xas);
                gxas.add(ga);
                return gxas;
            }
            return xas;
        }

        return ga;
    }

    // add the axis views recursively
    var as;
    as = CollectRowAxes.call(this, gridChom);
    if (as) this.addLeftAxis(as);
    as = CollectColAxes.call(this, gridChom);
    if (as) this.addBottomAxis(as);
};

YAHOO.lang.extend(GDC.charting.view.GridWithAxesView,
                  GDC.charting.view.ChartWithAxesView, {
});

})();
