var GDC = GDC || {};
GDC.charting = GDC.charting || {};
GDC.charting.layout = GDC.charting.layout || {};

/**
 * XUL-like layout abstraction.
 * @module layout
 */
GDC.charting.layout.Box = function() {
    this.shape = {x: 0, y: 0, w: 0, h: 0};
    this.bounds = {wmin: 0, hmin: 0};
    this.flex = 0;
};

YAHOO.lang.augmentObject(GDC.charting.layout.Box.prototype, {
    dump: function() {
        console.log("F %d [%d, %d] {%d, %d}",
                    this.flex,
                    this.shape.w, this.shape.h,
                    this.shape.x, this.shape.y);
    },

    reflow: function(w, h) {
        var s = this.shape;
        s.w = w;
        s.h = h;
        return null;
    //    return { w: w, h: h };
    }
});



/**
 * @class ContainerBox
 * @namespace GDC.charting.layout
 */
GDC.charting.layout.ContainerBox = function() {
    GDC.charting.layout.ContainerBox.superclass.constructor.call(this);

    this.children = [];
};

YAHOO.lang.extend(GDC.charting.layout.ContainerBox, GDC.charting.layout.Box, {
    add: function(e) {
        this.children.push(e);
    },

    dump: function() {
        GDC.charting.layout.ContainerBox.superclass.dump.call(this);

        console.group();
        this.children.forEach(function(e) {
            e.dump();
        });
        console.groupEnd();
    },

    reflow: function(w, h, ctx) {
        this.shape.w = w;
        this.shape.h = h;

        // just reflow the children's contents
        this.children.forEach(function(e) {
            e.reflow(e.shape.w, e.shape.h, ctx);
        });
    }
});

/**
 * @class VBox
 * @namespace GDC.charting.layout
 */
GDC.charting.layout.VBox = function() {
    GDC.charting.layout.VBox.superclass.constructor.call(this);
};

YAHOO.lang.extend(GDC.charting.layout.VBox, GDC.charting.layout.ContainerBox, {
    dump: function() {
        console.log("VBox");
        GDC.charting.layout.VBox.superclass.dump.call(this);
    },

    reflow: function(w, h, ctx) {
        function reflower(ctx) {
            var flex = 0;
            var flexChildren = [];
            var remainingSize = h;
            var ret = false;

            // calculate the flex sum and fixed height sum
            this.children.forEach(function(e) {
                if (e.flex) {
                    flex += e.flex;
                    flexChildren.push(e);
                } else {
                    var s = e.shape;
                    s.w = w;

                    var es = e.reflow(s.w, s.h, ctx);
                    if (es) {
                        s.w = es.w;
                        s.h = es.h;
                    }

                    remainingSize -= s.h;
                }
            });

            if (!flexChildren.length) {
                h -= remainingSize;
                return Math.round(remainingSize) !== 0;
            }

            // height adjustment caused by the bounds.hmin underflow
            var dh = 0;
            var remainingChildren = [];

            // set the height according to the flex factor
            flexChildren.forEach(function(e) {
                var s = e.shape;
                var h = (remainingSize * e.flex) / flex;

                // check the min bound
                if (h < e.bounds.hmin) {
                    dh += e.bounds.hmin - h;

                    s.w = w;
                    s.h = e.bounds.hmin;

                    var es = e.reflow(s.w, s.h, ctx);
                    if (es) {
                        s.w = es.w;
                        s.h = es.h;
                    }
                    if (s.w != w) {
                        w = s.w;
                        ret = true; // need to reflow again
                    }

                    remainingSize -= s.h;
                    flex -= e.flex;
                    return;
                }

                remainingChildren.push(e);
            });

            // set the height according to the flex factor
            remainingChildren.forEach(function(e) {
                var s = e.shape;
                s.w = w;
                s.h = (remainingSize * e.flex) / flex;

                var es = e.reflow(s.w, s.h, ctx);
                if (es) {
                    s.w = es.w;
                    s.h = es.h;
                }
                if (s.w != w) {
                    w = s.w;
                    ret = true; // need to reflow again
                }

                remainingSize -= s.h;
                flex -= e.flex;
            });

            return ret;
        }


        var ow = w, oh = h;
        //GDC.util.log("vbox.reflow [%d, %d]", w, h);

        this.shape.w = w;
        this.shape.h = h;

        // reflow the children sizes
        var loop;
        do {
            loop = reflower.call(this, ctx);
        } while(loop);

        // now reposition the children
        var yoff = 0;
        this.children.forEach(function(e) {
            e.shape.y = yoff; 
            yoff += e.shape.h;
        });

        return { w: w, h: h };
    }
});


/**
 * @class HBox
 * @namespace GDC.charting.layout
 */
GDC.charting.layout.HBox = function() {
    GDC.charting.layout.HBox.superclass.constructor.call(this);
};

YAHOO.lang.extend(GDC.charting.layout.HBox, GDC.charting.layout.ContainerBox, {
    dump: function() {
        console.log("HBox");
        GDC.charting.layout.HBox.superclass.dump.call(this);
    },

    reflow: function(w, h, ctx) {
        function reflower(ctx) {
            var flex = 0;
            var flexChildren = [];
            var remainingSize = w;
            var ret = false;

            // calculate the flex sum and fixed height sum
            this.children.forEach(function(e) {
                if (e.flex) {
                    flex += e.flex;
                    flexChildren.push(e);
                } else {
                    var s = e.shape;
                    s.h = h;

                    var es = e.reflow(s.w, s.h, ctx);
                    if (es) {
                        s.w = es.w;
                        s.h = es.h;
                    }

                    remainingSize -= s.w;
                }
            });

            if (!flexChildren.length) {
                w -= remainingSize;
                return Math.round(remainingSize) !== 0;
            }

            // height adjustment caused by the bounds.hmin underflow
            var dw = 0;
            var remainingChildren = [];

            // set the height according to the flex factor
            flexChildren.forEach(function(e) {
                var s = e.shape;
                var w = (remainingSize * e.flex) / flex;

                // check the min bound
                if (w < e.bounds.wmin) {
                    dw += e.bounds.wmin - w;

                    s.h = h;
                    s.w = e.bounds.wmin;

                    var es = e.reflow(s.w, s.w, ctx);
                    if (es) {
                        s.w = es.w;
                        s.h = es.h;
                    }
                    if (s.h != h) {
                        h = s.h;
                        ret = true; // need to reflow again
                    }

                    remainingSize -= s.w;
                    flex -= e.flex;
                    return;
                }

                remainingChildren.push(e);
            });

            // set the height according to the flex factor
            remainingChildren.forEach(function(e) {
                var s = e.shape;
                s.h = h;
                s.w = (remainingSize * e.flex) / flex;

                var es = e.reflow(s.w, s.h, ctx);
                if (es) {
                    s.w = es.w;
                    s.h = es.h;
                }
                if (s.h != h) {
                    h = s.h;
                    ret = true; // need to reflow again
                }

                remainingSize -= s.w;
                flex -= e.flex;
            });

            return ret;
        }


        var ow = w, oh = h;
        //GDC.util.log("hbox.reflow [%d, %d]", w, h);

        this.shape.w = w;
        this.shape.h = h;

        // reflow the children sizes
        var loop;
        do {
            loop = reflower.call(this, ctx);
        } while(loop);

        // now reposition the children
        var xoff = 0;
        this.children.forEach(function(e) {
            e.shape.x = xoff; 
            xoff += e.shape.w;
        });

        return { w: w, h: h };
    }
});


/**
 * @class Visual
 * @namespace GDC.charting.layout
 */
GDC.charting.layout.Visual = function() {
    GDC.charting.layout.Visual.superclass.constructor.call(this);
};

YAHOO.lang.extend(GDC.charting.layout.Visual, GDC.charting.layout.ContainerBox);
