dojo.require("dojo.colors"); // to have all the CSS3 color definitions 
dojo.require("dojox.gfx"); 

(function(){

var Box = GDC.charting.layout.Box;
var VBox = GDC.charting.layout.VBox;
var HBox = GDC.charting.layout.HBox;


GDC.charting.Renderer = function(surfaceElementName) {
    this.surfaceName = surfaceElementName;
};

YAHOO.lang.augmentObject(GDC.charting.Renderer.prototype, {
    createSurface: function(visual, w, h) {
        return new dojox.gfx.createSurface(
                      document.getElementById(this.surfaceName),
                      w, h);
    },

    render: function(s, visual) {
        this.renderBox(s, visual, visual.gfx);
    },

    renderContainerBox: function(g, box, gfx) {
        // check whether there are children to render
        if (!box.children.length) {
            return null;
        }

        if (gfx) {
            g = gfx;
        } else {
            g = g.createGroup();
        }

        var scope = this;
        box.children.forEach(function(e) {
            scope.renderBox(g, e, e.gfx);
        });

        var s = box.shape;
        if (s.x || s.y) {
            var m = dojox.gfx.matrix;
            g.applyTransform(m.translate(s.x,s.y));
        }

        box.gfx = g;
        return g;
    },

    renderBox: function(g, box, gfx) {
        // render using the object's method if available
        if (YAHOO.lang.isFunction(box.render)) {
            return box.render(g, gfx);
        }

        // ContainerBox rendering (all children)
        if (box instanceof GDC.charting.layout.ContainerBox) {
            return this.renderContainerBox(g, box, gfx);
        }
        return null;
    }
});



GDC.charting.view.Circle = function(color) {
    GDC.charting.view.Circle.superclass.constructor.call(this);
    this.color = color;
};

YAHOO.lang.extend(GDC.charting.view.Circle, GDC.charting.layout.Visual, {
    dump: function() {
        console.log("Circle: " + this.color);
        GDC.charting.view.Circle.superclass.dump.call(this);
    },
    
    render: function(g, gfx) {
        var s = this.shape;
        this.gfx = g.createEllipse({cx: s.x + s.w/2, cy: s.y + s.h/2,
                                    rx: s.w/2, ry: s.h/2})
                    .setFill(this.color);
        return this.gfx;
    }
});


var ch = new HBox();
var circle;
circle = new GDC.charting.view.Circle('lightblue');
circle.flex = 2;
ch.add(circle);
circle = new GDC.charting.view.Circle('lightblue');
circle.flex = 1;
ch.add(circle);
circle = new GDC.charting.view.Circle('lightblue');
circle.shape.w = 40;
ch.add(circle);

var c = new GDC.charting.view.ChartWithAxesView( ch );

circle = new GDC.charting.view.Circle('lightgreen');
circle.shape.w = 30;
c.addLeftAxis(circle);

var v = new VBox();
circle = new GDC.charting.view.Circle('lightgreen');
circle.flex = 2;
v.add(circle);
circle = new GDC.charting.view.Circle('lightskyblue');
circle.flex = 1;
v.add(circle);

v.shape.w = 40;
c.addLeftAxis(v);

circle = new GDC.charting.view.Circle('lightcoral');
circle.shape.h = 30;
c.addBottomAxis(circle);

//c.calculate(600, 300);
//c.dump();

//new GDC.charting.Renderer('chart').render(c);


(function(){

var Series = GDC.charting.data.Series;
var IteratingSeries = GDC.charting.data.IteratingSeries;
var ControlledSeries = GDC.charting.data.ControlledSeries;
var InspectedSeries = GDC.charting.data.InspectedSeries;
var IntsSeries = GDC.charting.data.IntsSeries;

var px = new InspectedSeries(new Series([1, 3, 2, 4, 5]));
var py = new InspectedSeries(new Series([1, 5, 7, 5, 5]));
var ps = new InspectedSeries(new Series([9, 5, 7, 5, 2]));

var rows = new InspectedSeries(new Series([1, 2]));
var cols = new InspectedSeries(new Series([1, 2, 3]));

//var px = new InspectedSeries(new Series([6, 6, 6, 6, 6]));
var smi = new IteratingSeries(px);
var smc = new ControlledSeries(new Series([
                new Series([1,2]),
                new Series([2,3]),
                new Series([1,3]),
                new Series([2,2]),
                new Series([3,1])
              ]), smi);
var sa = new InspectedSeries(new Series([9, 5]));

var sc = new GDC.charting.def.ScatteredChart();
sc.setLineStyle({width: 4, color: 'lightcoral'});
var sc = new GDC.charting.def.BarChart();
//var sc = new GDC.charting.def.Pie();

var xa = new GDC.charting.def.XAxis();
xa.setTitle('X Axis');
xa.setIncludeZero();
xa.setRange(px.props);
xa.setStyle({fill: 'gray', size: 17}); // FIXME: style object?
sc.addAxis(xa);

var ya = new GDC.charting.def.YAxis();
ya.setTitle('Y Axis');
ya.setIncludeZero();
ya.setRange(py.props);
ya.setStyle({fill: 'black', size: 16}); // FIXME: style object?
sc.addAxis(ya);

var sbc = new GDC.charting.def.StackedBarChart();
sbc.colors = ['blue', 'lightblue', 'lightskyblue', 'lightgreen', 'lightcoral'];
sbc.setSizeSeries(new GDC.charting.def.DiscreteSeries(smi, ya.getScaler()));
sbc.setStackedSizeSeries(new GDC.charting.def.DiscreteSeries(smc, new GDC.charting.axes.Scaler(new InspectedSeries(smc, smi).props)));
sbc.setStackedColorSeries(new GDC.charting.def.DiscreteSeries(sa, null));
sbc.addAxis(xa);
sbc.addAxis(ya);

// FIXME: setters?
sc.colors = ['blue', 'lightblue', 'lightskyblue', 'lightgreen', 'lightcoral'];
sc.markers = ['BAR', 'ELLIPSE', 'BAR', 'ELLIPSE', 'BAR'];

//sc.setMarkers( new GDC.charting.def.Ellipse() );
sc.setMarkers( new GDC.charting.def.Marker() );

//sc.setXSeries(new GDC.charting.def.ContinuousSeries(px, xa.getScaler()));
//sc.setYSeries(new GDC.charting.def.ContinuousSeries(py, ya.getScaler()));

sc.setSizeSeries(new GDC.charting.def.DiscreteSeries(ps, new GDC.charting.axes.Scaler(ps.props)));
sc.setColorSeries(new GDC.charting.def.DiscreteSeries(ps, null));
//sc.setMarkerSeries(new GDC.charting.def.DiscreteSeries(ps, null));

// display the two charts next to each other
var grr = new GDC.charting.def.Grid();
grr.setColumnSeries(new GDC.charting.def.DiscreteSeries(cols, new GDC.charting.axes.Scaler(cols.props)));
grr.setRowSeries(new GDC.charting.def.DiscreteSeries(rows, new GDC.charting.axes.Scaler(rows.props)));
grr.setCells(sc);
grr.addAxis(xa);
grr.addAxis(ya);

// display the two charts next to each other
var gr = new GDC.charting.def.Grid();
gr.setColumnSeries(new GDC.charting.def.DiscreteSeries(cols, new GDC.charting.axes.Scaler(cols.props)));
gr.setRowSeries(new GDC.charting.def.DiscreteSeries(rows, new GDC.charting.axes.Scaler(rows.props)));
gr.setCells(sc);
//gr.addAxis(xa);
//gr.addAxis(ya);

// display the two charts next to each other
var grid = new GDC.charting.def.Grid();
grid.setColumnSeries(new GDC.charting.def.DiscreteSeries(cols, new GDC.charting.axes.Scaler(cols.props)));
grid.setRowSeries(new GDC.charting.def.DiscreteSeries(rows, new GDC.charting.axes.Scaler(rows.props)));
grid.setCells(sbc);
grid.addAxis(xa);
grid.addAxis(ya);

// build the CHOM
var chom = GDC.charting.chom.def.getBuilder(grid).construct({});

// wrap with the axes views around
var c = new GDC.charting.view.GridWithAxesView(chom);
circle = new GDC.charting.view.Circle('lightgreen');
circle.shape.w = 30;
//c.addLeftAxis(circle);
circle = new GDC.charting.view.Circle('lightcoral');
circle.shape.h = 30;
//c.addBottomAxis(circle);

// calculate for the right width/height
//c = chom; c.reflow(600, 400);
var rr = new GDC.charting.Renderer('chart');
var rs = rr.createSurface(c, 600, 400);
c.calculate(600, 400, {gfx: new GDC.charting.xgfx.Gfx(rs)});
//c.dump();
rr.render(rs, c);

})();

})();

