var GDC = GDC || {};
GDC.charting.view = GDC.charting.view || {};

fontFamily = {family: "Tahoma"};

GDC.charting.view.TextBox = function(textShape) {
    GDC.charting.view.TextBox.superclass.constructor.call(this);

    this.textShape = textShape;
};

YAHOO.lang.extend(GDC.charting.view.TextBox, GDC.charting.layout.Visual, {
    reflow: function(w, h, ctx) {
        var s = this.shape;
        var ts = this.textShape;
        // measure the text and set the box size accordingly
        var e = ctx.gfx.getTextExtent(ts.text, ts.size);
        this.extent = e;
        //s.w = Math.max(e.w, w);
        //s.h = Math.max(e.h, h);

        // setting the width/height is essential here
        //
        // however I will not always need to readjust the width/height
        // I may just need to center the text inside the box as is.
        if (ts.angle) {
            s.w = e.h + e.h/3;
            s.h = h;
        } else {
            s.w = w;
            s.h = e.h + e.h/3;
        }
        return s;
    },

    render: function(g, gfx) {
        var ts = this.textShape;
        var e = this.extent;

        // center the text with respect to its extent
        var s = this.shape;

        this.gfx = g.createRect({x: s.x, y: s.y, width: s.w-1, height: s.h-1})
                    .setFill("rgba(200, 135, 130, 0.3)");

        var tts = {};
        if (ts.angle) {
            tts.x = s.x + s.w - e.h/4;
            tts.y = s.y + s.h/2 + e.w/2;
        } else {
            ts.x = s.x + s.w/2 - e.w/2;
            ts.y = s.y + s.h - e.h/4;
        }

        this.gfx = new GDC.charting.xgfx.Group(g).createText(ts)
                          .setFont({family:fontFamily.family, size:ts.size})
                          .setFill(ts.fill);
        var dg = this.gfx.getGfx();
        if (ts.angle) {
            dg.applyTransform(dojox.gfx.matrix.translate(tts.x, tts.y));
            dg.applyTransform(dojox.gfx.matrix.rotateg(this.gfx.angle));
        }
        return dg;
    }
});


GDC.charting.view.Axis = function(axisDef) {
    GDC.charting.view.Axis.superclass.constructor.call(this);

    this.def = axisDef;
};

YAHOO.lang.extend(GDC.charting.view.Axis, GDC.charting.layout.Visual, {
    calculate: function(pixels, tsize) {
        var a = this.def;

        function itCount(it) {
            var cnt = 0;
            while (it.hasNext()) { it.next(); cnt++; }
            return cnt;
        } 

        if (a.ticks) {
            a.range = a.range || {};
            a.range.min = 0;
            a.range.max = itCount( a.ticks.getIterator()); // FIXME: series length
        }
        var min = a.range.min;
        var max = a.range.max;

        // This is the actual axis range (the min/max metric values shown
        // on the axis). Normally one would want at least the closest
        // greater than axis range ticks to be visible.
        //
        // FIXME: this should only work for vertical axes
        var axisRange = this.getTickByFontSize( min, max, pixels, tsize);

        // if this is an attribute axis -> do not extend beyond the
        // designated min/max range
        if (a.ticks) {
            axisRange.min = min;
            axisRange.max = max;
            // number of ticks we can fit in
            axisRange.tick = Math.ceil((max-min) / (pixels / (tsize*2)));
            if (axisRange.tick < 1)
                axisRange.tick = 1;
        }

        // hmm, margins??
        ratio = pixels / (axisRange.max - axisRange.min);

        this.range = axisRange;

        // generate label texts
        this.labels = this.makeLabels( axisRange );

        // alternate line for horizontal labels
        this.labels.alternate = this.labels.maxWidth > (axisRange.tick*ratio) ? 1 : 0;
    },

    getTickByFontSize: function( min, max, availablePixels, fontSize ) {
        function digits(x) {
            return x !== 0 ? Math.floor(Math.log(Math.abs(x))/Math.LN10) : 1;
        }

        // number of digits in the metric tick labels
        var digs = digits(max);
        if ( min !== 0 ) digs = Math.min(digits(min), digs);
        var dim = Math.pow(10, digs);

        // tick distance (number of ticks) determined based
        // on the font size
        var dtick = 0;
        do {
            dim /= 10; // tick label rounding by magnitudes
            dtick = Math.floor(
                    Math.round(max-min) / (availablePixels/fontSize/3.4)
                    / dim) * dim;
        } while (dtick <= 0);

        var dmax = Math.floor(((max+dtick)) / dtick) * dtick;
        var dmin = Math.ceil(((min-dtick)) / dtick) * dtick - dtick/2;

        // if one of the local extremes is 0 then let the last
        // tick be 0 as well (no negative ticks for positive only ranges)
        if (min === 0) dmin = 0;
        if (max === 0) dmax = 0;

        return {min: dmin, max:dmax, tick:dtick};
    },

    makeLabels: function(ar) {
        var a = this.def;
        var tsize = a.style.size;
        var ggfx = this.ggfx;

        var ticktexts = [];
        var maxew = -1;

        var t, e;

        if (a.ticks) {
            var it = a.ticks.getIterator();
            var idx = 0;

            //for(var idx=ar.min; idx<=ar.max; idx+=ar.tick) 
            while(it.hasNext() ) {
                t = it.next();

                if (idx !== 0) {
                    var tickCnt = ar.tick;
                    while(--tickCnt > 0 && it.hasNext() ) {
                        t = it.next();
                        idx++;
                    }
                }

                e = ggfx.getTextExtent(t, tsize);

                // count it into the maximum axis label width
                maxew = Math.max(maxew, e.w);

                ticktexts.push({idx:idx, extent:e, text:t});
                idx++;
            }

            // push the past-end tick (with no text) to allow
            // 'center' alignment
            ticktexts.push({idx:idx});

            return {texts: ticktexts, maxWidth: maxew};
        }

        var from = ar.min;
        if (from < 0) from = 0;

        // 0 ... up
        for(idx=from; idx<=ar.max; idx+=ar.tick) {
            t = String( Math.round(idx * 1000) / 1000 );
            e = ggfx.getTextExtent(t, tsize);

            // count it into the maximum axis label width
            maxew = Math.max(maxew, e.w);

            ticktexts.push({idx:idx, extent:e, text:t});
        }

        // 0 ... down
        for(idx=-ar.tick; idx>=ar.min; idx-=ar.tick) {
            t = String( Math.round(idx * 1000) / 1000 );
            e = ggfx.getTextExtent(t, tsize);

            // count it into the maximum axis label width
            maxew = Math.max(maxew, e.w);

            ticktexts.push({idx:idx, extent:e, text:t});
        }

        return {texts: ticktexts, maxWidth: maxew};
    }
});


GDC.charting.view.LabeledXAxis = function(axisDef) {
    GDC.charting.view.LabeledXAxis.superclass.constructor.call(this, axisDef);

    this.layout = new GDC.charting.layout.VBox();
    // reflowing changes this.shape directly
 //FIXME: what the hell????   //this.shape = this.layout.shape;
    // add it as a child (for the ContainerBox.render() to work)
    this.add(this.layout); //this.layout.flex = 1;
    
    // add the axis and the title
    this.layout.add(new GDC.charting.view.XAxis(this.def));
    if (this.def.title) {
        this.layout.add(new GDC.charting.view.TextBox({
            text: this.def.title,
            size: this.def.style.size,
            fill: this.def.style.fill }));
    }
};

YAHOO.lang.extend(GDC.charting.view.LabeledXAxis, GDC.charting.view.Axis, {
    reflow: function(w, h, ctx) {
        var s = this.layout.reflow(w, h, ctx);
        return s;
    }
});


GDC.charting.view.LabeledYAxis = function(axisDef) {
    GDC.charting.view.LabeledYAxis.superclass.constructor.call(this, axisDef);

    this.layout = new GDC.charting.layout.HBox();
    // reflowing changes this.shape directly
 //FIXME: what the hell????   //this.shape = this.layout.shape;
    // add it as a child (for the ContainerBox.render() to work)
    this.add(this.layout); //this.layout.flex = 1;
    
    // add the title and the axis
    if (this.def.title) {
        this.layout.add(
            new GDC.charting.view.TextBox({
                text: this.def.title,
                angle: -90,
                size: this.def.style.size,
                fill: this.def.style.fill
            }));
    }
    this.layout.add(new GDC.charting.view.YAxis(this.def));
};

YAHOO.lang.extend(GDC.charting.view.LabeledYAxis, GDC.charting.view.Axis, {
    reflow: function(w, h, ctx) {
        return this.layout.reflow(w, h, ctx);
    }
});


GDC.charting.view.XAxis = function(axisDef) {
    GDC.charting.view.XAxis.superclass.constructor.call(this, axisDef);
};

YAHOO.lang.extend(GDC.charting.view.XAxis, GDC.charting.view.Axis, {
    reflow: function(w, h, ctx) {
        this.ggfx = ctx.gfx;
        this.calculate(w, this.def.style.size);

        var s = this.shape;
        var tsize = this.def.style.size;
        s.w = w;

        // FIXME rotation
        s.h = tsize + tsize/2; // tick labels + ticklen
        return s;
    },

    render: function(g, gfx) {
        var a = this.def;
        var tsize = a.style.size;
        var ticklen = tsize/4;
        var w = this.shape.w;

        var s = this.shape;
        g.createRect({x: s.x, y: s.y, width: s.w-1, height: s.h-1})
                    .setFill("rgba(135, 250, 206, 0.3)");

        var sx, e, text;
        var sy = 2;

        // axis line
        if (!a.ticks) {
            g.createPath()
                .moveTo(0, sy).lineTo(w-2, sy)
                .setStroke({color:a.style.fill, width:2.5});
        }

        // shift some more if alternating lines
        sy += this.labels.alternate * tsize;

        var line = this.labels.alternate;

        // right align the tick labels and set their stroke&fill
        var scope = this;
        this.labels.texts.forEach(function(t) {
            var gg, text;
            var e = t.extent;

            sx = a.scaler.map(t.idx, w);
            var s = { x: sx, y: sy };

            // tick
            if (!a.ticks) {
                g.createPath()
                    .moveTo(s.x, s.y).lineTo(s.x, s.y+ticklen)
                    .setStroke({color:a.style.fill, width:2.5});
            }

            if (!t.text) return;

            gg = g.createGroup();
            text = gg.createText({text:t.text})
                     .setFont({family:fontFamily.family, size:e.h});

            // tick label
            if (a.style.align === "center") {
                s.x += advance / 2;
            }
            if (a.style.rotate) {
                tr.rotate(text, a.style.rotate);

                var rot = Math.abs(a.style.rotate);
                var sinRot = Math.sin((rot / 180) * Math.PI);
                var cosRot = Math.cos((rot / 180) * Math.PI);
                if (a.style.rotate > 0) {
                    s.y -= scope.labels.maxWidth * sinRot + e.h/3;
                    s.x -= e.h/3; // FIXME should probably not be adjusted
                } else {
                    s.y -= (scope.labels.maxWidth - e.w) * sinRot + e.h/3;
                    s.x -= e.w * cosRot - e.h/3;
                }
            } else {
                s.x -= e.w/2; // align center wrt the tick

                line ^= scope.labels.alternate;
            }

            s.y += +ticklen/2 + tsize + tsize * line;
            gg.setTransform(dojox.gfx.matrix.translate(s.x, s.y));
            text.setFill(a.style.fill).setStroke({color:a.style.fill, width:0.5});
        });
        return g;
    }
});


GDC.charting.view.YAxis = function(axisDef) {
    GDC.charting.view.YAxis.superclass.constructor.call(this, axisDef);
};

YAHOO.lang.extend(GDC.charting.view.YAxis, GDC.charting.view.Axis, {
    reflow: function(w, h, ctx) {
        this.ggfx = ctx.gfx;
        this.calculate(h, this.def.style.size);

        var s = this.shape;
        var tsize = this.def.style.size;
        s.h = h;

        // FIXME rotation
        s.w = this.labels.maxWidth + tsize;
        return s;
    },

    render: function(g, gfx) {
        var a = this.def;
        var tsize = a.style.size;
        var ticklen = tsize/4;
        var h = this.shape.h;

        var s = this.shape;
        this.gfx = g.createRect({x: s.x, y: s.y, width: s.w-1, height: s.h-1})
                    .setFill("rgba(135, 206, 250, 0.3)");

        var sx = s.x;
        var sy = s.y;

        sx += this.shape.w - ticklen; // this.labels.maxWidth + tsize/2;

        // axis line
        if (!a.ticks) {
            g.createPath()
             .moveTo(sx+ticklen, 0).lineTo(sx+ticklen, h-2)// - a.scaler.map(this.range.min) + (this.range.tick*ratio)/3)
             .setStroke({color:a.style.fill, width:2.5});
        }

        // right align the tick labels and set their stroke&fill
        var scope = this;
        this.labels.texts.forEach(function(t) {
            var gg, text;
            var e = t.extent;

            var sy = h - a.scaler.map(t.idx, h);
            var s = { x: sx, y: sy };

            // tick
            if (!a.ticks) {
                g.createPath()
                 .moveTo(s.x, s.y).lineTo(s.x+ticklen, s.y)
                 .setStroke({color:a.style.fill, width:2.5});
            }

            if (!t.text) return;

            // tick label
            gg = g.createGroup();
            text = gg.createText({text:t.text})
                     .setFont({family:fontFamily.family, size:e.h});

            s.x -= e.h/2 + e.w; // align right
            s.y += e.h/4; // text baseline adjustments

            if (a.style.align === "center") {
                s.y -= advance/2;
            }
            if (a.style.rotate) {
                // for left-side axis (labels to the left of ticks)
                // rotate the text around its lower-right corner
                tr.move(text, e.w, 0);
                tr.rotate(text, a.style.rotate);
                tr.move(text, -e.w, 0);
            }
            gg.setTransform(dojox.gfx.matrix.translate(s.x, s.y));
            text.setFill(a.style.fill).setStroke({color:a.style.fill, width:0.5});
        });
    }
});
