/**
 * Dojox.gfx extension wrapper module.
 *
 * @module GDC.charting.xgfx
 */
GDC.charting.xgfx = GDC.charting.xgfx || {};


/**
 * Group wrapper used for further object construction
 *
 * @class Group
 * @module GDC.charting.xgfx
 */
GDC.charting.xgfx.Gfx = function(surface) {
    // create and hide the measurement group
    var measurement = surface.createGroup();
    measurement.getNode().style.visibility='hidden';

    this.measuredText = measurement.createText();
};


(function(){

/**
 * Get a dojox.gfx.Text extent (this set to the instance).
 */
function getExtent(text, height) {
    // TODO: dojox.gfx bug? where it tries to get the text node
    // not counting on the posibility of there not being one
    if (text === '') {
        return {w:0, h: height};
    }

    try {
        return {w: this.getTextWidth(), h: height};
    } catch(exception) {
        GDC.util.error("getTextWidth: %o", exception);
        return {w: -1, h: height};
    }
}
    
YAHOO.lang.augmentObject(GDC.charting.xgfx.Gfx.prototype, {
    /**
     * Get a text extent (basically bounding box in other words.
     */
    getTextExtent: function(text, size) {
        this.measuredText.setShape({text:text}).setFont({size:size});
        return getExtent.call(this.measuredText, text, size);
    },

    /**
     * Get bounding box out of a dojox.gfx.shape. There are some shapes
     * that simply do not implement the getBoundingBox() method. Therefore
     * we go DFS inside of such shape and get the bounding box from inside.
     * This is possible and perhaps even more accurate for markers.
     */
    getBBox: function(gfx) {
        var bb = gfx.getTransformedBoundingBox();
        if (!bb && gfx.children) {
            gfx.children.forEach(function(e) {
                ebb = getBBox(e);
                if (ebb) {
                    bb = ebb;
                }
            });
        }
        return bb;
    }
});

})();

/**
 * Group wrapper used for further object construction
 *
 * @class Group
 * @module GDC.charting.xgfx
 */
GDC.charting.xgfx.Group = function(group) {
    this.dgroup = group;
}

YAHOO.lang.augmentObject(GDC.charting.xgfx.Group.prototype, {
    createText: function(shape) {
        return new GDC.charting.xgfx.Text(this.dgroup, shape);
    },
    createTextArea: function(g, shape) {
        var t = new GDC.charting.xgfx.TextArea(this.dgroup, shape);
    }
});


/**
 * Text shape providing extent calculation (including rotation).
 * Does not rotate the text graphics though as it is not clear
 * around which point it should rotate.
 *
 * @class Text
 * @module GDC.charting.xgfx
 */
GDC.charting.xgfx.Text = function(group, shape) {
    this.dtext = group.createText();
    this.setShape(shape);
}

YAHOO.lang.augmentObject(GDC.charting.xgfx.Text.prototype, {
    setShape: function(s) {
        this.angle = s.angle;
        this.size = s.size;

        this.dtext.setShape(s);
        return this;
    },
    setStroke: function(s) {
        this.dtext.setStroke(s);
        return this;
    },
    setFill: function(f) {
        this.dtext.setFill(f);
        return this;
    },
    setFont: function(f) {
        this.dtext.setFont(f);
        return this;
    },

    getGfx: function() {
        return this.dtext;
    },
    getAngle: function() {
        return this.angle;
    },

    /**
     * Extent of the text graphics object.
     * w/h are the real, possible rotated extent
     * tw is the text width (graphics string advance)
     *
     * @method getExtent
     */
    getExtent: function() {
        var a = this.angle;
        var w, h = this.size;
        var tw = this.dtext.getTextWidth();
        if (a) {
            var sr = Math.sin((a / 180) * Math.PI);
            var cr = Math.cos((a / 180) * Math.PI);
            w = w * cr + h * sr;
            h = w * sr + h * cr;
        } else {
            w = tw;
        }
        return {w: w, h: h, tw: tw};
    }
});



GDC.charting.xgfx.TextArea = function(shape) {
    // FIXME: TODO
    //this.setShape(shape);
}

